Package: GloScope
Type: Package
Title: Population-level Representation on scRNA-Seq data
Version: 2.3.0
Authors@R: c(
  person("Elizabeth", "Purdom", email = "epurdom@stat.berkeley.edu", role = c("aut","cre")),
  person("William", "Torous", email = "wtorous@berkeley.edu",
            role = c("aut"),
	    comment = c(ORCID="0000-0001-5668-5510")),
  person("Hao", "Wang", email = "hao_wang@berkeley.edu",
            role = c("aut"),
            comment = c(ORCID="0000-0002-0749-474X")),
  person("Boying", "Gong", email = "boyinggong@berkeley.edu", role = "aut"))
Description: This package aims at representing and summarizing the entire single-cell profile of 
  a sample. It allows researchers to perform important bioinformatic analyses at the sample-level 
  such as visualization and quality control. The main functions Estimate sample distribution and calculate
  statistical divergence among samples, and visualize the distance matrix through MDS plots. 
BugReports: https://github.com/epurdom/GloScope/issues
License: Artistic-2.0
Encoding: UTF-8
Imports: utils, stats, MASS, mclust, ggplot2, RANN, FNN, BiocParallel,
        mvnfast, SingleCellExperiment, rlang, RColorBrewer, pheatmap,
        vegan, cluster, boot, permute
Depends: R (>= 4.4.0)
Suggests: BiocStyle, testthat (>= 3.0.0), knitr, rmarkdown,
        zellkonverter
VignetteBuilder: knitr
LazyData: false
biocViews: DataRepresentation, QualityControl, RNASeq, Sequencing,
        Software, SingleCell
RoxygenNote: 7.3.3
Config/testthat/edition: 3
git_url: https://git.bioconductor.org/packages/GloScope
git_branch: devel
git_last_commit: 15d93b8
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 22:54:52 UTC; biocbuild
Author: Elizabeth Purdom [aut, cre],
  William Torous [aut] (ORCID: <https://orcid.org/0000-0001-5668-5510>),
  Hao Wang [aut] (ORCID: <https://orcid.org/0000-0002-0749-474X>),
  Boying Gong [aut]
Maintainer: Elizabeth Purdom <epurdom@stat.berkeley.edu>

# Generated by roxygen2: do not edit by hand

export(bootCI)
export(bootGloscope)
export(getMetrics)
export(gloscope)
export(gloscopeProp)
export(plotCI)
export(plotHeatmap)
export(plotMDS)
importFrom(BiocParallel,SerialParam)
importFrom(BiocParallel,bplapply)
importFrom(BiocParallel,bpparam)
importFrom(FNN,KL.dist)
importFrom(MASS,isoMDS)
importFrom(RANN,nn2)
importFrom(RColorBrewer,brewer.pal)
importFrom(SingleCellExperiment,SingleCellExperiment)
importFrom(SingleCellExperiment,colData)
importFrom(SingleCellExperiment,reducedDim)
importFrom(boot,boot)
importFrom(boot,boot.ci)
importFrom(cluster,silhouette)
importFrom(ggplot2,aes)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_pointrange)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,position_dodge)
importFrom(grDevices,colorRampPalette)
importFrom(mclust,densityMclust)
importFrom(mvnfast,rmvn)
importFrom(permute,how)
importFrom(permute,permute)
importFrom(pheatmap,pheatmap)
importFrom(rlang,.data)
importFrom(stats,as.dist)
importFrom(stats,predict)
importFrom(stats,rmultinom)
importFrom(utils,combn)
importFrom(vegan,adonis2)
importFrom(vegan,anosim)

# GloScope Package

This repository implements the `GloScope` methodology for sample-level analysis of scRNA-Seq data. The details of this methodology can be found in the following pre-print:

[H. Wang, W. Torous, B. Gong, E. Purdom (2023).
Visualizing scRNA-Seq Data at Population Scale with GloScope. bioRxiv.](https://doi.org/10.1101/2023.05.29.542786)

A vignette demonstrating the basic usage of this package can be found in the [vignettes folder](https://github.com/epurdom/GloScope/tree/main/vignettes) of this repository.

## Installation

This package is available on Bioconductor, and the following R code can be used to install it.

```{r}
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("GloScope")
```

#' @title Plot confidence intervals
#'
#' @description This function creates a `ggplot` object that plots the
#'   confidence intervals created by `bootCI`
#'
#' @param ci_df A data.frame that is the output of \code{\link{bootCI}}
#' @param ci_df A data frame contains each sample's metadata. Note this is
#'   NOT at the cell-level.
#' @param color_by The column name or index in ci_df that should be used
#'   to color the confidence intervals by. 
#' @param group_by The column name or index in ci_df that should be used to
#'   determine how to group the confidence intervals. If missing all confidence
#'   intervals will be plotted in an order determined internally.
#' @param dodge_width value passed to `width` argument of
#'   \code{\link[ggplot2]{position_dodge}} if `group_by` variable is given.
#'   Controls separation between confidence intervals with the same grouping
#'   value.
#' @return A plot of sample-pair divergences with confidence intervals
#'
#' @examples
#' data(example_SCE_small)
#' sample_ids <- SingleCellExperiment::colData(example_SCE_small)$sample_id
#' # Run gloscope on first 10 PCA embeddings
#' # We use 'KNN' option for speed ('GMM' is slightly slower)
#' pca_embeddings <- SingleCellExperiment::reducedDim(example_SCE_small,"PCA")
#' pca_embeddings_subset <- pca_embeddings[,seq_len(10)] # select the first 10 PCs
#' dist_result <- gloscope(pca_embeddings_subset, sample_ids,
#'    dens="KNN",
#'    BPPARAM = BiocParallel::SerialParam(RNGseed=2))
#' # make a per-sample metadata
#' sample_metadata <- as.data.frame(unique(SingleCellExperiment::colData(example_SCE_small)[,c(1,2)]))
#' # make another variable
#' sample_metadata$fakeGroup<-c(rep(c("A","B"),each=2),"A")
#' manyboot<-bootCI(dist_result,
#'   sample_metadata,"sample_id",
#'   metrics=c("anosim","silhouette"),group_vars=c("phenotype","fakeGroup"),R=20)
#' plotCI(manyboot,group_by="metric",color_by="grouping")
#' 
#' @export
#' @importFrom ggplot2 aes
#' @importFrom ggplot2 geom_pointrange
#' @importFrom ggplot2 ggplot
#' @importFrom ggplot2 position_dodge

plotCI<-function(ci_df,color_by,group_by,dodge_width=.5){

  if(!all(c('distance','grouping','metric','statistic','lower','upper') %in% names(ci_df)))
    stop("names of ci_df must match the expected output of bootCI")
  whCols<-c("metric","distance","grouping")
  listCols<-paste(whCols,collapse=",")
  whMulti<-sapply(ci_df[,whCols,drop=FALSE],function(x){nlevels(factor(x))>1})
  if(!missing(color_by)){
    if(!c(color_by) %in% whCols) stop("color_by must be one of: ",listCols)
    if(!color_by %in% whCols[whMulti]) stop("color_by must have more than one distinct value")
  }
  if(!missing(group_by)){
    if(!c(group_by) %in% whCols) stop("group_by must be one of: ",listCols)
    if(!group_by %in% whCols[whMulti]) stop("group_by must have more than one distinct value")
  }
  if(missing(group_by)){
    pos="identity"
    if(sum(whMulti)>0){
      ci_df$ConfInt <- apply(ci_df[,whCols][,whMulti,drop=FALSE],1,paste,collapse=":")
      ci_df$ConfInt <- factor(ci_df$ConfInt)
    }
    else{ci_df$ConfInt<-""}
    group_by="ConfInt"
    if(missing(color_by)){
      ggp <- ggplot(ci_df,
                    aes(x = .data[[group_by]], y = .data$statistic)
      )
    }
    else{
      ggp <- ggplot(ci_df,
                      aes(x = .data[[group_by]], y = .data$statistic,
                          color = .data[[color_by]])
        )
    }
  }
  else{
    if(sum(whMulti)>2) stop("Cannot accurate group confidence intervals if they vary by more than two of:",listCols)
    pos <- position_dodge(width = dodge_width)
    if(missing(color_by)){
      ggp <- ggplot(ci_df,
                    aes(x = .data[[group_by]], y = .data$statistic)
      )
    }
    else{
      ggp <- ggplot(ci_df,
                    aes(x = .data[[group_by]], y = .data$statistic,
                        color = .data[[color_by]])
      )
    }
  }  
  ggp <- ggp + geom_pointrange(aes(ymin = .data$lower, ymax = .data$upper),
                  position = pos, size = 0.9)
  return(ggp)
}

#' @title Calculate divergence between all sample pairs' cell type proportion
#'
#' @description This function calculates a matrix of pairwise divergences
#'   between input samples' cell type proportion.
#'
#' @param cell_sample_ids a vector of the samples IDs each cell comes from. Length
#'   must match the number of element in `cell_type_ids`
#' @param cell_type_ids a vector of user defined cell type
#' @param ep an numeric value added to the summary counts. Default ep = 0 means
#'   nothing will be added.
#' @param dist_metric metric to calculate the divergence between samples.
#' @return clusprop_dist a symmetric matrix of divergences 
#' @details Options for `dist_metric` are as follows: "KL" calculates the
#'   symmetric-KL divergence. "JS" calculates the Jenson-Shannon distance. "TV"
#'   calculates the total variation distance.
#' @examples
#' # Bring in small example data of single cell embeddings
#' data(example_SCE_small)
#' sample_id <- SingleCellExperiment::colData(example_SCE_small)$sample_id 
#' cluster_id <- SingleCellExperiment::colData(example_SCE_small)$cluster_id 
#' dist_result <- gloscopeProp(sample_id, cluster_id, ep = 0.5, 
#'                                    dist_metric = "KL")
#' dist_result
#' @importFrom utils combn
#' @rdname gloscopeProp
#' @export

gloscopeProp <- function(cell_sample_ids, cell_type_ids,
                                ep = 0, dist_metric = c("KL", "JS", "TV")){
  if(length(cell_sample_ids)!=length(cell_type_ids)){
    stop("Lengths of cell id and cell type are not equal!")
  }
  cell_sample_ids <- as.character(cell_sample_ids)
  cell_type_ids <- as.character(cell_type_ids)
  
  cluster_table <- table(cell_sample_ids, cell_type_ids)
  clusprop <- matrix(cluster_table, ncol = ncol(cluster_table), 
                     dimnames = dimnames(cluster_table))
  if(dist_metric %in% c("KL","JS")){
    if(sum(clusprop==0)>0 & ep == 0){
      warning("There are cell-types in some samples having 0 counts! You may get invalid results with this distance type. Please consider setting ep>0, e.g 0.5.")
    }else if (sum(clusprop==0)>0 & ep != 0){
      message(paste0("There are cell-types in some samples having 0 counts. An amount ", ep, " has been added to all counts."))
    }
    clusprop<- clusprop+ep
    
  }
  clusprop <- t(apply(clusprop, 1, function(x) x/sum(x)))
  
  unique_sample_ids <- rownames(clusprop)
  sample_pairs <- utils::combn(unique_sample_ids, 2)
  # Convert patient pairs to a list for BiocParallel::bplapply
  patient_pair_list <- lapply(seq_len(ncol(sample_pairs)), function(i) sample_pairs[,i])
  divergence_list <- lapply(patient_pair_list,
                            function(w){ .calc_prop(prop1 = clusprop[w[1],], 
                                                  prop2 = clusprop[w[2],], 
                                                  dist_metric = dist_metric)})
  
  # Convert pair-wise distances to a symmetric distance matrix
  divergence_vec <- unlist(divergence_list)
  divergence_matrix <- matrix(0, ncol = length(unique_sample_ids),
                              nrow = length(unique_sample_ids))
  rownames(divergence_matrix) <- unique_sample_ids
  colnames(divergence_matrix) <- unique_sample_ids
  divergence_matrix[lower.tri(divergence_matrix, diag=FALSE)] <- divergence_vec
  divergence_matrix[upper.tri(divergence_matrix)] <- t(divergence_matrix)[upper.tri(divergence_matrix)]
  
  return(divergence_matrix)
}

# Helper function to obtain KL divergence using cell type proportion

#' @title Helper function to obtain the KL distance using cell type proportion
#'
#' @description The helper function `.calc_prop` calculates the KL divergence from
#'   two set of proportions.
#'
#' @param prop1 sample1's cluster proportion
#' @param prop2 sample2's cluster proportion
#' @param dist_metric distance metric to calculate the distance. One of
#'   c("KL","JS")
#' @return A single value contains the  KL divergence value calculated for the 
#'    2 samples' cluster proportion.
#'
#' @noRd
.calc_prop <- function(prop1, prop2, dist_metric){
  if(dist_metric == "KL"){
    out <-  sum(prop1*(log(prop1) - log(prop2))) +
      sum(prop2*(log(prop2) - log(prop1)))
  }else if(dist_metric == "JS"){
    out <-  1/2* sum(prop1*(log(prop1) - log(1/2*prop1 + 1/2*prop2))) +
      1/2* sum(prop2*(log(prop2) - log(1/2*prop1 + 1/2*prop2)))
  }
  else if(dist_metric == "TV"){
    out <-  1/2* sum(abs(prop1-prop2))
  }
  return(out)
}

#' @title Calculate density for each sample based on reduced dimension embedding
#'
#' @description The `calc_dens` function fits separate multivariate densities to
#'   each sample. As input, the function expects a named list with names
#'   corresponding to sample IDs, and elements holding a matrix or data.frame of
#'   dimensionality reduced cells from the samples. Each matrix-type will have
#'   rows corresponding to each cell and columns corresponding to the cells
#'   projection into a latent dimension. The k-nearest neighbour algorithm does
#'   density and distance estimation in a single step, and for that `dens`
#'   specification this function simply returns the input embedding matrices
#'   (see R/calc_dist.R).
#'
#' @param df_list A list containing each samples' dimension reduction embedding,
#'   i.e. a list of matrices/data.frames. It is REQUIRED that each of these
#'   matrices have column names.
#' @param dens method used to estimate density, options are GMM (Gaussian
#'   mixture model) and KNN (k-nearest Neighbor)
#' @param k number of k nearest neighbour for KNN density estimation
#' @param num_components a vector of integers for the number of components to
#'   fit GMMS to
#' @param GMM_params optional GMM parameters
#' @param BPPARAM BiocParallel parameters
#' @return mod_list: a list of length number of samples, contains the estimated
#'   density for each sample
#'
#' @importFrom BiocParallel bplapply
#' @importFrom BiocParallel bpparam
#' @importFrom mclust densityMclust
#' @noRd
.calc_dens <- function(df_list, dens = c("GMM","KNN"), k, num_components,
                       GMM_params, BPPARAM){
  dens<-match.arg(dens)
  if(dens == "GMM"){
    # run (in parallel) GMM density fitting with `mclust::densityMclust`
    if ("G" %in% names(GMM_params)){
      stop("G cannot be specified in `GMM_params`. This is specified by `num_components` instead.")
    }
    # `mclust::densityMclust()` expects column names when invoked through
    # `do.call()` (required for R >= 4.5, otherwise it errors while assigning
    # dimnames).  
    hasNoNames<-sapply(df_list,function(x){is.null(colnames(x))})
    if(any(hasNoNames)) stop("df_list must be list of matrices with columnnames for fitting the GMM density. The function gloscope has a check for this, so if you are getting this error you are calling `.calc_dens` directy. `.calc_dens` is not intended as a user-facing function.")
    # Normalise the matrices once up front to avoid repeatedly
    # copying large objects inside the parallel loop.
    # df_list <- lapply(df_list, function(mat){
    #   if (is.null(colnames(mat))) {
    #     colnames(mat) <- paste0("V", seq_len(ncol(mat)))
    #   }
    #   mat
    # })
    
    mod_list <- BiocParallel::bplapply(df_list, function(z){
      GMM_params$data <- z
      GMM_params$G <- get_gmm_num_components_vec(nrow(z),num_components)
      fit_model <- do.call(mclust::densityMclust,GMM_params)
      return(fit_model)
    }, BPPARAM=BPPARAM)
  }else if(dens == "KNN"){
    # The KNN algorithm takes the embedding coordinates as input and does not require density estimation
    mod_list <- df_list
  }
  
  return(mod_list)
}

#' @title Draw samples from a fit GMM for Monte Carlo approximation
#'
#' @description The helper function `.sample_mclust` draws `r` samples from a
#'   GMM fit to a sample's latent embedding. This is used in `.calc_KL()` to
#'   compute a Monte Carlo approximation of the symmetric KL divergence between
#'   sample pairs.
#'
#' @param mclust_mod An `mclustDensity` object of a sample's estimated density
#' @param r Number of Monte Carolosimulations to generate, default = 10000.
#' @return A matrix contains the symmetrised KL divergence value calculated for
#'   each pair of samples.
#'
#' @importFrom stats rmultinom
#' @importFrom mvnfast rmvn
#' @noRd

# Helper function to sample from a fit `mclust` GMM
.sample_mclust <- function(mclust_mod, r){
    p <- table(mclust_mod$classification)/mclust_mod$n
    z <- stats::rmultinom(1, size = r, prob = p)[,1]
    non_zero <- which(z!=0) # clusters with zero samples raise a sampling error
    samples <- lapply(non_zero, function(i){
        mvnfast::rmvn(z[i], mu = mclust_mod$parameters$mean[, i],
        sigma = mclust_mod$parameters$variance$sigma[, , i])
    })
    samples <- do.call("rbind", samples)
    return(samples)
}

#' @title Helper function to query nearest neighbour distances
#'
#' @description This helper function `.knn_query` gets the distance from each
#'   cell in a query matrix to its k-th nearest neighbour in a target matrix.
#'   This is used to approximate various statistical divergences in
#'   `R/calc_dist.R`.
#'
#' @param df_list A named list with each sample's reduced dimension embedding
#' @param query The name or index of the sample whose query matrix is used to
#'   compute distances from
#' @param input The name or index of the sample whose matrix is used to find the
#'   k-th NN
#' @param KNN_params a list of arguments for the RANN::nn2 funciton, including `k`
#' @return A matrix contains the symmetrised KL divergence value calculated for
#'   each pair of samples.
#'
#' @importFrom RANN nn2
#' @noRd
.knn_query <- function(df_list, query, input, KNN_params){
    KNN_params$data <- df_list[[input]]
    KNN_params$query <- df_list[[query]]
    knnq_dist <- do.call(RANN::nn2,KNN_params)$nn.dists[,KNN_params$k]
    return(knnq_dist)
}



#' @title Helper function to pick number of GMM components
#'
#' @description `mclust::densityMclust` will raise an error if the 
#'   number of components vector `G` contains values at least as 
#'   large as the number of samples. If the user sets `num_components`
#'   with invalid entries, this function replaces them with n-1 and
#'   raises a warning.
#'
#' @param num_obs The number of observations to fit
#' @param num_components The user's requested vector of components
#' @return A valid vector of GMM number of components to fit
#'                                                                           
#' @noRd  
get_gmm_num_components_vec <- function(num_obs, num_components){
    if(any(num_components > (num_obs - 1))){
    warning(paste0("Unable to fit a GMM with ",
            paste(num_components[num_components > (num_obs - 1)], collapse = ", "),
        " components to a sample with ", num_obs, " cells.",
        " Replacing those parameters with ", num_obs - 1, " components instead."))
    }
    num_components[num_components > (num_obs - 1)] <- (num_obs - 1)
    return(unique(num_components)) # avoid duplicates
}

.as_full_matrix <- function(x) {
  m <- as.matrix(x)
  # ensure square, named, and symmetric indexability
  if (is.null(rownames(m)) && !is.null(colnames(m))) rownames(m) <- colnames(m)
  if (is.null(colnames(m)) && !is.null(rownames(m))) colnames(m) <- rownames(m)
  m
}

.testDistMeta<-function(dist_mat,metadata_df,sample_id){
  if(!c(sample_id) %in% names(metadata_df)) stop("sample_id does not define a variable in metadata_df")
  if(nrow(dist_mat)!= nrow(metadata_df)){
    stop("Not consistent patient number. Make sure your
            distance matrix and meta info have the same patient number.")
  }
  if(length(intersect(rownames(dist_mat), metadata_df[, sample_id]))!= nrow(dist_mat)){
    stop("Not consistent patient IDs. Make sure your
            distance matrix and meta info have the same patients.")
  }
  if(!(identical(rownames(dist_mat), metadata_df[, sample_id]))){
    metadata_df <- metadata_df[match(rownames(dist_mat), metadata_df[,sample_id]),,drop=FALSE]
  }
  return(metadata_df)
  
}

#' @title Plot a heatmap of the GloScope represenation
#'
#' @description This function creates a heatmap of the given GloScope divergence
#'   matrix.
#'
#' @param dist_mat The divergence matrix output of `gloscope()`. Should be a
#'   symmetric, square matrix.
#' @param metadata_df A data frame contains each sample's metadata. Note this is
#'   NOT at the cell-level, and should have the same number of rows as dist_mat.
#' @param sample_id The column name or index in metadata_df that contains the
#'   sample ID. This is for ensuring alignment between the dist_mat and the
#'   metadata_df. The rownames of dist_mat are expected to match the sample_id
#'   values.
#' @param color_by A vector of column names or indices in metadata_df that should be used
#'   to color/annotate the samples. 
#' @param which_side One of "columns","rows", or "both", indicating whether the
#'   annotation of the samples in `color_by` should be on the rows, columns, or
#'   on both.
#' @param ... parameters passed to \code{\link[pheatmap]{pheatmap}}.
#' @return Invisibly returns the output of \code{\link[pheatmap]{pheatmap}}
#'
#' @details The function is a wrapper function to \code{\link[pheatmap]{pheatmap}}.
#'   `color_by` is used to create subset of the `metadata_df` to pass to
#'   `annotation_col` (if `which_side="columns"`) or `annotation_row` (if
#'   `which_side="rows"`). If `which_side="both"`, then it is passed to both,
#'   and `annotation_names_row` argument is set to `FALSE`, suppressing labeling
#'   both the columns and rows (which user can thus not override). All other
#'   arguments to \code{\link[pheatmap]{pheatmap}} can be passed directly by the user
#' @seealso \code{\link[pheatmap]{pheatmap}}
#' @examples
#' data(example_SCE_small)
#' sample_ids <- SingleCellExperiment::colData(example_SCE_small)$sample_id
#' # Run gloscope on first 10 PCA embeddings
#' # We use 'KNN' option for speed ('GMM' is slightly slower)
#' pca_embeddings <- SingleCellExperiment::reducedDim(example_SCE_small,"PCA")
#' pca_embeddings_subset <- pca_embeddings[,seq_len(10)] # select the first 10 PCs
#' dist_result <- gloscope(pca_embeddings_subset, sample_ids,
#'    dens="KNN",
#'    BPPARAM = BiocParallel::SerialParam(RNGseed=2))
#' # make a per-sample metadata
#' sample_metadata <- as.data.frame(unique(SingleCellExperiment::colData(example_SCE_small)[,c(1,2)]))
#' plotHeatmap(dist_mat = dist_result, metadata_df = sample_metadata ,
#' sample_id="sample_id", color_by="phenotype")
#' # Pass additional options to pheatmap to control colors of groups
#' library(RColorBrewer)
#' plotHeatmap(dist_mat = dist_result, metadata_df = sample_metadata ,
#' sample_id="sample_id", color_by="phenotype", which_side="both", 
#' annotation_colors=list(phenotype = c(Covid = "magenta", Healthy = "white")), 
#' color = colorRampPalette(brewer.pal(9, "PuBuGn"))(100))

#' @importFrom pheatmap pheatmap
#' @importFrom RColorBrewer brewer.pal
#' @importFrom grDevices colorRampPalette
#' @importFrom stats as.dist
#' @export
#'
plotHeatmap <- function(dist_mat, metadata_df, sample_id, color_by, 
                        which_side=c("columns","rows","both"),
                        ...){
    which_side<-match.arg(which_side)
    metadata_df<-.testDistMeta(dist_mat,metadata_df,sample_id)
    
  # Create a data frame for the annotations
  if(!missing(color_by)){
    rownames(metadata_df)<-metadata_df[,sample_id]
    annotcol <- metadata_df[,color_by,drop=FALSE]
    if(which_side=="columns"){
      invisible(pheatmap(
        mat = as.dist(dist_mat),
        annotation_col = annotcol,
        ...)
      )
    }
    else if(which_side=="rows"){
      invisible(pheatmap(
        mat = as.dist(dist_mat),
        annotation_row = annotcol,
        ...)
      )
    }
    else if(which_side=="both"){
      invisible(pheatmap(
        mat = as.dist(dist_mat),
        annotation_row = annotcol,
        annotation_col = annotcol,
        annotation_names_row=FALSE,
        ...)
      )
    }
  }
  else{
    invisible(pheatmap(mat = as.dist(dist_mat),...))
  }

}

data(example_SCE)

sample_ids <- SingleCellExperiment::colData(example_SCE)$sample_id
set.seed(2)
subsample_patients <- sample(unique(sample_ids),4)
subsample_metadata <- as.data.frame(SingleCellExperiment::colData(example_SCE)[sample_ids %in% subsample_patients,])
subsample_data <- SingleCellExperiment::reducedDim(example_SCE,"PCA")[sample_ids %in% subsample_patients,]
subsample_data_subset <- subsample_data[,1:10] # Pick the first 10 PCs

undersized_patient <- subsample_patients[1]
set.seed(2)
sample_drop_indices <- sample(which(subsample_metadata$sample_id==undersized_patient),(500-49),replace=FALSE)
undersized_metadata <- subsample_metadata[-sample_drop_indices,]
undersized_data <- subsample_data[-sample_drop_indices,]
undersized_data_subset <- undersized_data[,1:10]

#for patient level tests need a gloscope matrix to test on
expect_silent(dist_mat <- gloscope(embedding_matrix=subsample_data_subset, cell_sample_ids=subsample_metadata$sample_id,dens="KNN"))
pat_info <- unique(subsample_metadata[,c(1,2)])
pat_info$group <- c("A","A","B","B")

pat_info_full<-as.data.frame(unique(SingleCellExperiment::colData(example_SCE)[,c("sample_id","phenotype")]))
pat_info_full$group<-rep(c("A","B"),each=10)
#create fake dist_mat of 20x20
n<-nrow(pat_info_full)
rvalues<-round(abs(rnorm(n*(n-1)/2)),3)
dist_mat_full<-matrix(0,nrow=n,ncol=n)
dist_mat_full[upper.tri(dist_mat_full)]<-rvalues
dist_mat_full[lower.tri(dist_mat_full)]<-t(dist_mat_full)[lower.tri(dist_mat_full)]
rownames(dist_mat_full)<-colnames(dist_mat_full)<-pat_info_full$sample_id

#' @rdname getMetrics
#' @details The function `bootGloscope` is a wrapper to the
#'   \code{\link[boot]{boot}} function for creating bootstraps of one of the
#'   metrics calculated by `getMetrics`. Most users will probably prefer `bootCI`
#' @param R number of bootstrap replicates. See \code{\link[boot]{boot}}.
#' @param ... arguments passed to \code{\link[boot]{boot}}
#' @return `bootGloscope` returns an object of class `boot` created by \code{\link[boot]{boot}}.
#' @seealso \code{\link[boot]{boot}}
#' @examples
#' # single bootstrap of anosim
#' bootout<-bootGloscope(dist_result,sample_metadata,"sample_id",
#'   metric="anosim",group_var="phenotype")
#' #work with the boot object using functions in boot package:
#' library(boot)
#' print(bootout)
#' boot.ci(bootout)
#' 
#' @importFrom boot boot
#' @export
#' 
bootGloscope<-function(dist_mat, metadata_df, metrics="anosim",sample_id, group_vars,R=1000,...){
  metadata_df<-.testDistMeta(dist_mat,metadata_df,sample_id)
  if(!c(group_vars) %in% names(metadata_df)) stop("group_vars does not define a variable in metadata_df")
  combinedData<-cbind(dist_mat,metadata_df)
  nsamples<-ncol(dist_mat)
  if(length(group_vars)>1) stop("group_vars must be single character value")
  if(length(metrics)>1) stop("metrics must be single character value")
  bootfun<-function(df,i){#x are ids
    d<-df[,1:nsamples]
    m<-df[,(nsamples+1):ncol(df)]
    getMetrics(d[i,i], m[i,], metrics=metrics,sample_id=sample_id, 
                group_vars=group_vars,checkData=FALSE, permuteTest=FALSE)$statistic
  }
  return(boot::boot(data=combinedData,statistic=bootfun,R=R,...))
    
}

#' @rdname getMetrics
#' @order 2
#' @param ci_type Single character value. The type of confidence interval to
#'   compute. Passed to argument `type` in \code{\link[boot]{boot.ci}}
#' @param ci_conf Scalar value between 0 and 1. The confidence level requested.
#'   Passed to argument `conf` in \code{\link[boot]{boot.ci}}.
#' @details `bootCI` is a wrapper function to
#'   \code{\link[boot]{boot.ci}}. `boot.ci` can be called directly on the output
#'   of `bootGloscope`. The main advantage of `bootCI` is to calculate
#'   bootstrap CI over multiple choices of metrics, variables, and/or distance
#'   matrices. Unlike `boot.ci`, `bootCI` does not allow different
#'   choices of confidence interval types or levels, so `ci_type` and `ci_level`
#'   must be of length 1. For this kind of multiplicity, call `boot.ci` directly
#'   on the output of `bootGloscope`.
#' @return `bootCI` creates a data frame containing the statistic for
#'   each combination of metric and grouping variable with columns with the
#'   upper and lower bounds of the requested confidence intervals
#' \itemize{
#'   \item metric 
#'   \item grouping
#'   \item statistic
#'   \item lower
#'   \item upper
#' }
#' @seealso \code{\link[boot]{boot.ci}}
#' @examples
#' # calculate many bootstraps -- for speed up we set R ridiculously low
#' manyboot<-bootCI(list("Distance 1"=dist_result,"Another distance"=dist_result),
#'   sample_metadata,"sample_id",
#'   metrics=c("anosim","silhouette"),group_vars=c("phenotype","grouping"),R=20)
#' 
#' @importFrom boot boot.ci
#' @export
bootCI<-function(dist_mat, metadata_df, metrics="anosim",sample_id, group_vars,R=1000,ci_type=c("perc","norm","basic", "stud",  "bca"),ci_conf=0.95,...){
  ci_type=match.arg(ci_type)
  if(length(ci_conf)!=1) stop("only single value of ci_conf is allowed")
  if(!inherits(dist_mat,"list")){
    dist_mat<-list(dist_mat)
  }
  if(inherits(dist_mat,"list") & is.null(names(dist_mat))){
    names(dist_mat)<-paste("Distance",1:length(dist_mat),sep="")
  }
  if(length(ci_type)!=1) stop("only a single type of confidence interval can be done with this wrapper function")
  comb<-expand.grid(distName=names(dist_mat),g=group_vars,m=metrics)
  perComboFunction<-function(distName,g,m){
    distName<-as.character(distName)
    g<-as.character(g)
    m<-as.character(m)
    out<-bootGloscope(dist_mat[[distName]],metadata_df,sample_id, metrics=m,group_vars=g,...)
    out.ci<-boot.ci(out,type=ci_type,conf=ci_conf)
    ci_type<-match.arg(ci_type,names(out.ci)) #make sure have the full name
    nCols<-ncol(out.ci[[ci_type]])
    return(data.frame(distance=distName,grouping=g,metric=m,statistic=out.ci$t0,lower=out.ci[[ci_type]][,nCols-1],upper=out.ci[[ci_type]][,nCols]))
    
  }
  do.call("rbind",do.call("mapply",c(list(FUN=perComboFunction,SIMPLIFY=FALSE),comb)))
}

library(testthat)
all_models <- c("EII","VII","EEI","VEI","EVI","VVI","EEE","VEE","EVE","VVE","EEV","VEV","EVV","VVV")
test_that("gloscope warnings for small cell counts", {
	expect_error(gloscope(subsample_data_subset, c(subsample_metadata$sample_id,NA)),
		regexp="The number of cells in the embedding matrix does not match the number of sample labels.",fixed=TRUE)
	sample_size_warnings <- capture_warnings(na_div_matrix <- gloscope(undersized_data_subset,undersized_metadata$sample_id,dens="KNN"))
	expect_match(sample_size_warnings,"^The following samples have fewer than 500 cells.*",all = FALSE)
	expect_match(sample_size_warnings,"^The following samples have fewer than the minimum of 50 cells*",all = FALSE)
	# confirm the NA divergences are added last for samples with less than k cells
	expect_equal(unname(colnames(na_div_matrix))[4],c("CV0234"))
	expect_equal(unname(rownames(na_div_matrix))[4],c("CV0234"))
	expect_equal(all(is.na(na_div_matrix[-4,-4])),FALSE)
	expect_equal(all(is.na(na_div_matrix["CV0234",])),TRUE)
	expect_equal(all(is.na(na_div_matrix[,"CV0234"])),TRUE)
	
	# confirm a warning when the user requests more GMM components than cells
	# Case when there are some valid component numbers specified
	num_component_warnings1 <- capture_warnings(updated_clusters <- get_gmm_num_components_vec(49,c(1, 49)))
	expect_match(num_component_warnings1,"^Unable to fit a GMM with 49 components to a sample with 49 cells. Replacing those parameters with 48 components instead.*",all = FALSE)
	expect_equal(updated_clusters,c(1,48))
	# Case when no valid component numbers specified
	num_component_warnings2 <- capture_warnings(small_gloscope <- gloscope(undersized_data_subset,
	    undersized_metadata$sample_id,dens="GMM",num_components=c(50),GMM_params = list(modelNames=all_models,verbose=FALSE,plot=FALSE)))
	expect_match(num_component_warnings2,"^Unable to fit a GMM with 50 components to a sample with 49 cells. Replacing those parameters with 48 components instead.*",all = FALSE)
	expect_equal(dim(small_gloscope),c(4,4)) # Confirm a matrix is returned
})

test_that("gloscope works with KNN",{
  expect_silent(temp_knn<-gloscope(subsample_data_subset,subsample_metadata$sample_id,
      dens = "KNN", dist_metric = "KL",BPPARAM = BiocParallel::SerialParam(RNGseed=2)))
  #test dimensions
  expect_equal(dim(temp_knn),c(4,4))
  #test distances the same as in the past
  knn_expected_answer<-c(2.369535,4.478759,2.215369,3.203009,3.225507,1.520805) #answer got the first time
  expect_equal(round(temp_knn[upper.tri(temp_knn)],6),knn_expected_answer)
  #test diag zero
  expect_equal(unname(diag(temp_knn)),rep(0,4))
  #test row names/colnames
  expect_equal(unname(sort(colnames(temp_knn))),sort(as.character(unique(subsample_metadata$sample_id))))
  expect_equal(unname(sort(rownames(temp_knn))),sort(as.character(unique(subsample_metadata$sample_id))))
  # test symmetry
  expect_equal(isSymmetric(temp_knn),TRUE)
  
  
})

test_that("gloscope works with GMM",{
  expect_silent(temp_gmm<-gloscope(subsample_data_subset,subsample_metadata$sample_id,
      dens = "GMM", GMM_params = list(modelNames=all_models,verbose=FALSE,plot=FALSE),
      num_components = seq_len(9), dist_metric = "KL",BPPARAM = BiocParallel::SerialParam(RNGseed=1)))
  #test dimensions
  expect_equal(dim(temp_gmm),c(4,4))
  #test distances the same as in the past
  gmm_expected_answer<-round(c(14.220048,13.089617,18.073184,10.097857,11.156613,10.425697),6) #answer got the first time
  expect_equal(round(temp_gmm[upper.tri(temp_gmm)],6),gmm_expected_answer)
  #test diag zero
  expect_equal(unname(diag(temp_gmm)),rep(0,4))
  #test row names/colnames
  expect_equal(unname(sort(colnames(temp_gmm))),sort(as.character(unique(subsample_metadata$sample_id))))
  expect_equal(unname(sort(rownames(temp_gmm))),sort(as.character(unique(subsample_metadata$sample_id))))
  # test symmetry
  expect_equal(isSymmetric(temp_gmm),TRUE)

})

test_that("gloscope works with variations in embedding properties",{
  #check works with data.frames
  exdf<-data.frame(subsample_data_subset)
  expect_silent(gloscope(exdf,subsample_metadata$sample_id,
                         dens = "KNN", 
                         num_components = 3:4, dist_metric = "KL",BPPARAM = BiocParallel::SerialParam(RNGseed=1)))
  expect_silent(gloscope(exdf,subsample_metadata$sample_id,
                         dens = "GMM", 
                         num_components = 3:4, dist_metric = "KL",BPPARAM = BiocParallel::SerialParam(RNGseed=1)))
  
  
  #check what if no column names
  stripped<-subsample_data_subset
  colnames(stripped)<-NULL
  expect_silent(gloscope(stripped,subsample_metadata$sample_id,
                                   dens = "KNN", 
                                   num_components = 3:4, dist_metric = "KL",BPPARAM = BiocParallel::SerialParam(RNGseed=1)))
  expect_silent(gloscope(stripped,subsample_metadata$sample_id,
                                   dens = "GMM", 
                                   num_components = 3:4, dist_metric = "KL",BPPARAM = BiocParallel::SerialParam(RNGseed=1)))
})

test_that("different random seeds give different GMM results",{
  expect_false(all(gloscope(subsample_data_subset,subsample_metadata$sample_id,
                                   dens = "GMM", dist_metric = "KL",BPPARAM = BiocParallel::SerialParam(RNGseed=2)) ==
                 gloscope(subsample_data_subset,subsample_metadata$sample_id,
                          dens = "GMM", dist_metric = "KL",BPPARAM = BiocParallel::SerialParam(RNGseed=1))))
})



test_that("GMM density fitting returns `densityMclust` objects",{
  sample_ids <- subsample_metadata$sample_id
  # the following `lapply` creates the necessary input data structure for this fn.
  embeddings_list <- lapply(unique(sample_ids),function(x){subsample_data_subset[(sample_ids==x),]})
  names(embeddings_list) <- unique(sample_ids)
  expect_silent(gmm_density_list <- .calc_dens(embeddings_list, dens = "GMM", num_components=seq_len(9),
    GMM_params = list(plot=FALSE,verbose=FALSE), BPPARAM = BiocParallel::SerialParam(RNGseed = 2))
  )
  # confirm that each element is in the output is a `densityMclust` object
  expect_true(all(unlist(lapply(gmm_density_list,function(x){class(x)[1]=="densityMclust"}))))
})

test_that("JS divergences are properly implemented",{
  # From Toy Examples 1 and 2 from "On Accuracy of PDF Divergence Estimators
  # and Their Applicability to Representative Data Sampling" by Budka et al. (2011)

  # Example 1
  set.seed(2)
  s1 <- mvnfast::rmvn(10000,mu=c(Dim1=0,Dim2=0),sigma=diag(2),kpnames=TRUE)
  s2 <- mvnfast::rmvn(10000,mu=c(Dim1=0.5,Dim2=-0.5),sigma= matrix(c(0.5,0.1,0.1,0.3),2,2),kpnames=TRUE)
  df_list_1 <- list(s1,s2)
  
  #check KNN
  js_knn_1_expected <- 0.17
  expect_silent(mod_list_knn_1 <- .calc_dens(df_list_1, dens="KNN"))
  expect_silent(js_knn_1 <- .calc_JS (mod_list_knn_1, df_list_1, 1, 2,
    dens = "KNN", KNN_params = list(k=50)))
  expect_equal(round(js_knn_1,2),js_knn_1_expected)

  #check GMM
  js_gmm_1_expected <- 0.18
  expect_silent(mod_list_gmm_1 <- .calc_dens(df_list_1, dens="GMM", num_components = seq_len(9),
    GMM_params = list(plot=FALSE,verbose=FALSE),BPPARAM = BiocParallel::SerialParam(RNGseed = 2)))
  expect_silent(js_gmm_1 <- .calc_JS (mod_list_gmm_1, df_list_1, 1, 2, dens = "GMM", r = 10000))
  expect_equal(round(js_gmm_1,2),js_gmm_1_expected)

  # Example 2
  set.seed(2)
  s1 <- mvnfast::rmvn(10000,mu=c(Dim1=0,Dim2=0),sigma= matrix(c(1,0,0,0.1),2,2),kpnames=TRUE)
  s2 <- mvnfast::rmvn(10000,mu=c(Dim1=0,Dim2=0),sigma= matrix(c(0.1,0,0,1),2,2),kpnames=TRUE)
  df_list_2 <- list(s1,s2)

  #check KNN
  js_knn_2_expected <- 0.32
  mod_list_knn_2 <- .calc_dens(df_list_1, dens="KNN")
  js_knn_2 <- .calc_JS (mod_list_knn_2, df_list_2, 1, 2, dens = "KNN", KNN_params = list(k=50))
  expect_equal(round(js_knn_2,2),js_knn_2_expected)

  #check GMM
  js_gmm_2_expected <- 0.33
  mod_list_gmm_2 <- .calc_dens(df_list_2, dens="GMM", num_components = seq_len(9),
    GMM_params = list(plot=FALSE,verbose=FALSE),BPPARAM = BiocParallel::SerialParam(RNGseed = 2))
  js_gmm_2 <- .calc_JS (mod_list_gmm_2, df_list_2, 1, 2, dens = "GMM", r = 10000)
  expect_equal(round(js_gmm_2,2),js_gmm_2_expected)
})

test_that("the sKL divergences are properly implemented",{
  # From Toy Examples 1 and 2 from "On Accuracy of PDF Divergence Estimators
  # and Their Applicability to Representative Data Sampling" by Budka et al. (2011)

  # Example 1
  set.seed(2)
  s1 <- mvnfast::rmvn(10000,mu=c(Dim1=0,Dim2=0),sigma=diag(2),kpnames=TRUE)
  s2 <- mvnfast::rmvn(10000,mu=c(Dim1=0.5,Dim2=-0.5),sigma= matrix(c(0.5,0.1,0.1,0.3),2,2),kpnames=TRUE)
  df_list_1 <- list(s1,s2)

  kl_knn_1_expected <- 1.45
  mod_list_knn_1 <- .calc_dens(df_list_1, dens="KNN")
  kl_knn_1 <- .calc_kl(mod_list_knn_1, df_list_1, 1, 2,
    dens="KNN", KNN_params=list(k=50))
  expect_equal(round(kl_knn_1,2),kl_knn_1_expected)

  #check with and without colnames
  kl_gmm_1_expected <- 1.84
  mod_list_gmm_1 <- .calc_dens(df_list_1, dens="GMM", num_components = seq_len(9),
    GMM_params = list(plot=FALSE,verbose=FALSE),BPPARAM = BiocParallel::SerialParam(RNGseed = 2))
  kl_gmm_1 <- .calc_kl (mod_list_gmm_1, df_list_1, 1, 2, dens = "GMM")
  expect_equal(round(kl_gmm_1,2),kl_gmm_1_expected)

  # Example 2
  set.seed(2)
  s1 <- mvnfast::rmvn(10000,mu=c(Dim1=0,Dim2=0),sigma= matrix(c(1,0,0,0.1),2,2),kpnames=TRUE)
  s2 <- mvnfast::rmvn(10000,mu=c(Dim1=0,Dim2=0),sigma= matrix(c(0.1,0,0,1),2,2),kpnames=TRUE)
  df_list_2 <- list(s1,s2)

  kl_knn_2_expected <- 3.28
  mod_list_knn_2 <- .calc_dens(df_list_2, dens="KNN")
  kl_knn_2 <- .calc_kl (mod_list_knn_2, df_list_2, 1, 2, dens = "KNN",KNN_params = list(k=50))
  expect_equal(round(kl_knn_2,2),kl_knn_2_expected)

  kl_gmm_2_expected <- 4.05
  mod_list_gmm_2 <- .calc_dens(df_list_2, dens="GMM", num_components = seq_len(9),
    GMM_params = list(plot=FALSE,verbose=FALSE),BPPARAM = BiocParallel::SerialParam(RNGseed = 2))
  kl_gmm_2 <- .calc_kl (mod_list_gmm_2, df_list_2, 1, 2, dens = "GMM")
  expect_equal(round(kl_gmm_2,2),kl_gmm_2_expected)
})

test_that("Divergences are properly computed with GloScope inputs and KNN",{
  sample_ids <- subsample_metadata$sample_id
  embeddings_list <- lapply(unique(sample_ids),function(x){subsample_data_subset[(sample_ids==x),]})
  names(embeddings_list) <- unique(sample_ids)
  knn_density_list <- .calc_dens(embeddings_list, dens = "KNN")
  sample_pairs <- utils::combn(unique(sample_ids), 2)
  patient_pair_list <- lapply(seq_len(ncol(sample_pairs)), function(i) sample_pairs[,i])
  w <- as.character(patient_pair_list[[6]]) # pick one unit
  w_div_kl_expected <- 1.52
  w_div_kl <- .calc_kl(knn_density_list, embeddings_list, w[1], w[2],
    dens = "KNN", KNN_params = list(k = 50))
  w_div_js_expected <- 0.04
  w_div_js <- .calc_JS(knn_density_list, embeddings_list, w[1], w[2], 
    dens = "KNN", KNN_params = list(k = 50))
  expect_equal(round(w_div_kl,2),w_div_kl_expected)
  expect_equal(round(w_div_js,2),w_div_js_expected)
})


test_that("Divergences are properly computed with GloScope inputs and GMM",{
  sample_ids <- subsample_metadata$sample_id
  embeddings_list <- lapply(unique(sample_ids),function(x){subsample_data_subset[(sample_ids==x),]})
  names(embeddings_list) <- unique(sample_ids)
  gmm_density_list <- .calc_dens(embeddings_list, dens = "GMM",
    GMM_params = list(plot=FALSE,verbose=FALSE),num_components = seq_len(9), BPPARAM = BiocParallel::SerialParam(RNGseed = 2))
  sample_pairs <- utils::combn(unique(sample_ids), 2)
  patient_pair_list <- lapply(seq_len(ncol(sample_pairs)), function(i) sample_pairs[,i])
  w <- as.character(patient_pair_list[[2]]) # pick one unit
  w_div_kl_expected <- 5.06
  set.seed(2)
  w_div_kl <- .calc_kl(gmm_density_list, embeddings_list, w[1], w[2], dens = "GMM")
  w_div_js_expected <- 0.50
  set.seed(2)
  w_div_js <- .calc_JS(gmm_density_list, embeddings_list, w[1], w[2], dens = "GMM", r = 10000)
  expect_equal(round(w_div_kl,2),w_div_kl_expected)
  expect_equal(round(w_div_js,2),w_div_js_expected)
})

test_that("Divergences using cell type works properly",{
  sample_ids <- subsample_metadata$sample_id
  celltype <- subsample_metadata$cluster_id
  expect_warning(gloscopeProp(sample_ids,celltype, dist_metric= "KL"), "There are cell-types in some samples having 0 counts! You may get invalid results with this distance type.")
  expect_message(fix_div_matrix <-gloscopeProp(sample_ids,celltype, dist_metric= "KL",ep=0.5), "There are cell-types in some samples having 0 counts. An amount 0.5 has been added to all counts.")
  
  expect_equal(isSymmetric(fix_div_matrix),TRUE)
  short_sample_ids <- sample_ids[1:100]
  expect_error(gloscopeProp(short_sample_ids, celltype),
               regexp="Lengths of cell id and cell type are not equal!",fixed=TRUE)

#check other distances work:
  expect_message(gloscopeProp(sample_ids,celltype, dist_metric= "JS",ep=0.5), "There are cell-types in some samples having 0 counts. An amount 0.5 has been added to all counts.")
  expect_silent(gloscopeProp(sample_ids,celltype, dist_metric= "TV",ep=0.5))
  
})

test_that("gloscope warnings for user-specified models", {
    expect_error(gloscope(subsample_data_subset,
        subsample_metadata$sample_id,dens="GMM",num_components=c(50),
        GMM_params = list(G=c(50),verbose=FALSE,plot=FALSE)),
            regexp="G cannot be specified in `GMM_params`. This is specified by `num_components` instead.",fixed=TRUE)
    expect_error(gloscope(subsample_data_subset,
                          subsample_metadata$sample_id,dens="KNN",k=50,
        KNN_params = list(k=50)),
            regexp="k cannot be specified in `KNN_params`. This is specified by `k` instead.",fixed=TRUE)
})



test_that("getMetrics works",{
  ## if add new metrics, this has to change
  nmetrics_possible<-3
  result<-c("anosim"=0.25,"adonis2"=1.53,"silhouette"=0.07)
  expect_silent(test1<-getMetrics(dist_mat,metadata_df=pat_info, sample_id="sample_id", group_vars="phenotype"))
  expect_equal(dim(test1),c(nmetrics_possible,3))
  expect_equal(round(test1$statistic,2),unname(result))
  
  expect_silent(test2<-getMetrics(dist_mat,metadata_df=pat_info, sample_id="sample_id", group_vars=c("phenotype","group")))
  expect_equal(dim(test2),c(nmetrics_possible*2,3))
  expect_silent(test3<-getMetrics(dist_mat,metadata_df=pat_info, metrics="anosim",sample_id="sample_id", group_vars="phenotype"))
  expect_equal(dim(test3),c(1,3))

  #test permutations
  expect_no_error(test4<-getMetrics(dist_mat,metadata_df=pat_info, sample_id="sample_id", group_vars="phenotype",permuteTest=TRUE))
  expect_equal(dim(test4),c(nmetrics_possible,4))
  expect_equal(round(test4$statistic,2),unname(result))
  #test bootstrap
  expect_silent(bootout<-bootGloscope(dist_mat_full,metadata_df=pat_info_full,R=50,sample_id="sample_id", group_var=c("group")))
  expect_silent(bootCI(dist_mat_full,pat_info_full,sample_id="sample_id",
                              metric=c("anosim","silhouette"),group_var="phenotype",R=50,ci_type = "perc")
  )
  expect_silent(bootCI(dist_mat_full,pat_info_full,sample_id="sample_id",
                                          metric=c("anosim"),group_var="phenotype",R=50,ci_type = "norm")
  )
  expect_silent(bootCIout<-bootCI(list("Dist 1"=dist_mat_full, "New Dist"=dist_mat_full),pat_info_full,sample_id="sample_id",
                               metric=c("anosim","silhouette"),group_var="phenotype",R=50,ci_type = "perc"))
  expect_error(bootCI(list("Dist 1"=dist_mat_full, "New Dist"=dist_mat),pat_info_full,sample_id="sample_id",
                                metric=c("anosim"),group_var="phenotype",R=50,ci_type = "perc"), "Not consistent patient number")
  expect_error(bootCI(dist_mat_full,pat_info_full,sample_id="sample_id",
                                          metric=c("anosim"),group_var="phenotype",R=50,ci_type = "normal")
  )
  expect_silent(ggci<-plotCI(bootCIout))
  expect_silent(print(ggci))
  expect_silent(ggci<-plotCI(bootCIout,group_by="distance",color_by="metric"))
  expect_silent(print(ggci))
  expect_silent(ggci<-plotCI(bootCIout,group_by="metric",color_by="distance"))
  expect_silent(print(ggci))
  expect_error(plotCI(bootCIout,group_by="metric",color_by="grouping"),"color_by must have more than one distinct value")
  expect_error(plotCI(bootCIout,group_by="grouping",color_by="metric"),"group_by must have more than one distinct value")
  
})
test_that("plotMDS works with output",{
  expect_silent(mds_result <- plotMDS(dist_mat = dist_mat,
                                      metadata_df=pat_info, sample_id="sample_id", color_by="phenotype", k = 2))
  expect_silent(print(mds_result$plot))
  
  expect_silent(mds_result <- plotMDS(dist_mat = dist_mat,
                                      metadata_df=pat_info, sample_id="sample_id", shape_by="phenotype", k = 2))
  expect_silent(print(mds_result$plot))
  
  expect_silent(mds_result <- plotMDS(dist_mat = dist_mat,
                                      metadata_df=pat_info, sample_id="sample_id", shape_by="phenotype", color_by="phenotype", k = 2))
  expect_silent(print(mds_result$plot))
  
  expect_silent(mds_result <- plotMDS(dist_mat = dist_mat,
                                      metadata_df=pat_info, sample_id="sample_id", k = 2))
  expect_silent(print(mds_result$plot))
  
})

test_that("plotHeatmap works with output",{
  expect_silent(plotHeatmap(dist_mat = dist_mat,
                            metadata_df=pat_info, sample_id="sample_id", color_by="phenotype"))
  expect_silent(plotHeatmap(dist_mat = dist_mat,
                            metadata_df=pat_info, sample_id="sample_id", color_by="phenotype",
                            which_side="rows"))
  expect_silent(plotHeatmap(dist_mat = dist_mat,
                            metadata_df=pat_info, sample_id="sample_id", color_by="phenotype",
                            which_side="both"))
  ## Check passing other args to pheatmap
  expect_silent(plotHeatmap(dist_mat = dist_mat,metadata_df=pat_info, sample_id="sample_id", 
                            color_by=c("phenotype","phenotype"),
                            color = colorRampPalette(rev(brewer.pal(9, "RdYlBu")))(100)))
  
})

#' @title Calculate metrics on separation in distance matrix due to grouping variable
#'
#' @description These functions are wrappers for calculating common metrics for
#'   the amount of separation in a distance matrix due to a grouping (factor)
#'   variable and creating bootstrap confidence intervals and permutation tests.
#' 
#' @order 1
#' 
#' @param dist_mat The divergence matrix output of `gloscope()`. Should be a
#'   symmetric, square matrix. For `bootCI` the argument can be a list
#'   of distance matrices.
#' @param metadata_df A data frame contains each sample's metadata. Note this is
#'   NOT at the cell-level, and should have the same number of rows as dist_mat.
#' @param sample_id The column name or index in metadata_df that contains the
#'   sample ID. This is for ensuring alignment between the dist_mat and the
#'   metadata_df. The rownames of dist_mat are expected to match the sample_id
#'   values.
#' @param metrics vector of statistics to calculate. For `bootstrap_gloscope`
#'   must be single value.
#' @param group_vars vector of names of grouping variables in metadata_df for
#'   which to calculate metrics. For `bootstrap_gloscope` must be single value.
#' @param checkData Whether to check whether dist_mat, metadata_df, and
#'   sample_id match, for example in terms of dimensions and rownames. Mainly
#'   used internally.
#' @param permuteTest whether to run permutation tests on each of the metrics
#' @param permutations if `permuteTest=TRUE`, an integer value defines the
#'   number of permutations. Can also except output of
#'   \code{\link[permute]{how}} for more fine control of the permutation
#'   mechanisms.
#' @return `getMetrics` creates a data frame containing the statistic for each
#'   combination of metric and grouping variable with columns
#' \itemize{
#'   \item metric 
#'   \item grouping
#'   \item statistic
#'   \item pval (if `permuteTest=TRUE`)
#' }
#'
#' @details The function `getMetrics` is a simple wrapper for calculating statistics that
#'   summarize the difference between distances within and between groupings. If
#'   the variable defined by group_var does not have at least two groupings, the
#'   function will return a NA. 
#' @details The options "anosim" and "adonis2" are wrappers to the functions of
#'   that name in the package `vegan`; we have turned off the permutation
#'   testing option of those functions. The functions in `vegan` have greater
#'   capability, and in particular \code{\link[vegan]{adonis2}} has capability
#'   to handle more complicated testing paradigms than a simple grouping factor.
#'   Permutation tests for these statistics are handled by the functions of
#'   `vegan`.
#' @details The option "silhouette" calls the function of that name from the
#'   package `cluster` and calculates the silhouette width of each group and
#'   then averages them across groups. The permutation test is coded making use
#'   of the package `permute`, similar to `vegan`, so that control of the
#'   permutation mechanism is possible in the same way.
#' @seealso \code{\link[vegan]{anosim}}, \code{\link[vegan]{adonis2}}, 
#'  \code{\link[cluster]{silhouette}}, \code{\link[permute]{how}}
#' @examples
#' data(example_SCE_small)
#' sample_ids <- SingleCellExperiment::colData(example_SCE_small)$sample_id
#' # Run gloscope on first 10 PCA embeddings
#' # We use 'KNN' option for speed ('GMM' is slightly slower)
#' pca_embeddings <- SingleCellExperiment::reducedDim(example_SCE_small,"PCA")
#' pca_embeddings_subset <- pca_embeddings[,seq_len(10)] # select the first 10 PCs
#' dist_result <- gloscope(pca_embeddings_subset, sample_ids,
#'    dens="KNN",
#'    BPPARAM = BiocParallel::SerialParam(RNGseed=2))
#' # make a per-sample metadata
#' sample_metadata <- as.data.frame(unique(SingleCellExperiment::colData(example_SCE_small)[,c(1,2)]))
#' # make another variable
#' sample_metadata$grouping<-c(rep(c("A","B"),each=2),"A")
#' getMetrics(dist_result,metadata_df=sample_metadata, sample_id="sample_id", 
#'   group_vars="phenotype")
#' # run permutation tests:
#' getMetrics(dist_result,metadata_df=sample_metadata, sample_id="sample_id", 
#'   group_vars=c("phenotype","grouping"), permuteTest=TRUE)
#'   
#' @importFrom cluster silhouette
#' @importFrom vegan anosim
#' @importFrom permute how 
#' @importFrom permute permute
#' @importFrom vegan adonis2
#' @export
#'
getMetrics <- function(dist_mat, metadata_df, metrics=c("anosim","adonis2","silhouette"),
                        sample_id, group_vars, checkData=TRUE, permuteTest=FALSE,
                        permutations=100) {
  metrics<-match.arg(metrics,several.ok = TRUE)
  if(any(!c(sample_id,group_vars) %in% names(metadata_df))) stop("sample_id or group_vars does not define a variable in metadata_df")
  # test that at least 2 levels of each factor
  testnlevels<-sapply(group_vars,function(x){nlevels(factor(x))>1})
  # turn into dist for methods; align order using labels
  m <- .as_full_matrix(dist_mat)
  labs <- rownames(m)
  stopifnot(!is.null(labs))
  if(checkData) metadata_df<-.testDistMeta(dist_mat,metadata_df,sample_id)
  metadata_df[,sample_id] <- as.character(metadata_df[,sample_id])
  d <- as.dist(m)
  
  # to turn off permutation calculations:
  if(!permuteTest){
    permControl<-permute::how(
      minperm=0, maxperm=2,nperm=0
    )
  }else{
    permControl=permutations
  }
  combinedData<-cbind(dist_mat,metadata_df)
  nsamples<-ncol(dist_mat)
  
  getanosim<-function(varname){
    f<-factor(metadata_df[,varname])
    if(nlevels(f)<2) return(NA)
    else{
      out<-vegan::anosim(d,f ,permutations=permControl)
      if(!permuteTest) return(out$statistic)
      else return(c(statistic=out$statistic, pval=out$signif))
    }
  }
  getadonis2<-function(varname){
    f<-factor(metadata_df[,varname])
    if(nlevels(f)<2) return(NA)
    else{
      out<-vegan::adonis2(d~f,permutations=permControl)
      if(!permuteTest) return(out$F[1])
      else{ return(c(statistic=out$F[1],pval=out$`Pr(>F)`[1]))}
    }      
  }
  
  getsil<-function(varname){
    if(nlevels(factor(metadata_df[,varname]))<2) return(NA)
    else{
      meansil<-function(data,ind){
        dd<-data[,1:nsamples]
        m<-data[,(nsamples+1):ncol(data)]
        fac<-factor(m[,varname])
        sil<-cluster::silhouette(as.integer(factor(fac)), dd)
        return(mean(sil[,"sil_width"]))
      }
      if(!permuteTest){
        return(meansil(combinedData,ind=1:nsamples))
      }
      else{
        if(!inherits(permControl,"how")) permControl<-how(nperm=permControl)
        return(.permGeneric(data=combinedData,nobs=nsamples,statistic=meansil,control=permControl))
      }
    }
  }
  
  combs<-expand.grid(metric=metrics,grouping=group_vars)
  out<-apply(as.matrix(combs),1,function(x){
    metric_x<-as.character(x[[1]])
    var_x<-as.character(x[[2]])
    if(metric_x=="anosim") return(getanosim(var_x))
    if(metric_x=="adonis2") return(getadonis2(var_x))
    if(metric_x=="silhouette") return(getsil(var_x))
  })
  if(!permuteTest) combs$statistic<-out
  else combs<-cbind(combs,t(out))
  return(combs)
}


.permGeneric<-function(data,nobs,statistic,control){
  ## generate randomisation distribution 
  stat.obs<-statistic(data,ind=1:nobs)
  stat.permu<-numeric(length = control$nperm)
  for(i in seq_along(control$nperm)){
    ## return a permutation
    want <- permute::permute(i, nobs, control=control)
    ## calculate permute stat
    stat.permu[i] <-  statistic(data,ind=want)
  }
  ## pval from permutation test
  pval <- sum(abs(c(stat.permu,stat.obs)) >= abs(stat.obs))/(control$nperm+1)
  ## return value
  return(c(statistic = stat.obs, pval = pval))
}

#' @title Calculate statistical divergence between all sample pairs
#'
#' @description This function calculates a matrix of pairwise divergences
#'   between input samples of single cell data.
#'
#' @param embedding_matrix a matrix or data.frame of latent embeddings with rows
#'   corresponding to cells and columns to dimensions
#' @param cell_sample_ids a vector of the samples IDs each cell comes from. Length
#'   must match the number of rows in `embedding_matrix`
#' @param dens the density estimation. One of c("GMM","KNN")
#' @param dist_metric distance metric to calculate the distance. One of
#'   c("KL","JS")
#' @param r number of Monte Carlo simulations to generate
#' @param num_components a vector of integers for the number of components to
#'   fit GMMs to, default is c(5,10,15,20)
#' @param k number of nearest neighbours for KNN density estimation, default k =
#'   50.
#' @param GMM_params optional mclust parameters, default is to restrict the fit
#'   model to only VVE
#' @param KNN_params optional arguments for either `FNN:KL.dist` (KL) or
#'   `RANN::nn2` (JS), default is NULL
#' @param BPPARAM BiocParallel parameters, default is running in serial. Set
#'   random seed with `RNGseed` argument
#' @param prefit_density a named list of pre-fit `densityMclust` objects for
#'   each sample, default is NULL
#' @param return_density return the GMM parameter list or not (if applicable),
#'   default is FALSE
#' @return A matrix containing the pairwise divergence or distance between all
#'   pairs of samples
#'
#' @examples
#' # Bring in small example data of single cell embeddings
#' data(example_SCE_small)
#' sample_ids <- SingleCellExperiment::colData(example_SCE_small)$sample_id 
#' pca_embeddings <- SingleCellExperiment::reducedDim(example_SCE_small,"PCA")
#' # Run gloscope on first 10 PCA embeddings
#' # We use 'KNN' option for speed ('GMM' is slightly slower)
#' pca_embeddings_subset <- pca_embeddings[,seq_len(10)] # select the first 10 PCs
#' dist_result <- gloscope(pca_embeddings_subset, sample_ids,
#'          dens="KNN", BPPARAM = BiocParallel::SerialParam(RNGseed=2))
#' dist_result
#'
#' @importFrom BiocParallel bplapply
#' @importFrom BiocParallel SerialParam
#' @importFrom utils combn
#' @rdname gloscope
#' @export

gloscope <- function(embedding_matrix, cell_sample_ids,
                dens = c("GMM","KNN"), dist_metric = c("KL","JS"),
                r = 10000, num_components = c(5,10,15,20), k = 50,
                GMM_params = list(modelNames = c("VVE"),verbose=FALSE,plot=FALSE),
                KNN_params = NULL,
                BPPARAM = BiocParallel::SerialParam(),
                prefit_density = NULL, return_density = FALSE){
    dens<-match.arg(dens)
    dist_metric<-match.arg(dist_metric)
    # Input safety check
    if(length(cell_sample_ids)!=nrow(embedding_matrix)){
        stop("The number of cells in the embedding matrix does not match the number of sample labels.")
    }

    # Extract the unique sample IDs as characters
    cell_sample_ids <- as.character(cell_sample_ids)
    unique_sample_ids <- unique(cell_sample_ids)
    names(unique_sample_ids) <- unique_sample_ids

    if(is.null(colnames(embedding_matrix))){
      # GMM fitting now requires column names for matrix
      colnames(embedding_matrix)<-paste("Dim",seq_len(ncol(embedding_matrix)),sep="")
    }
    # Results may be unreliable for samples with less than 500 cells
    # We raise a warning denoting samples for which this is the case
    MIN_CELLS <- 500
    cells_per_sample <- vapply(unique_sample_ids,
        function(x){nrow(embedding_matrix[cell_sample_ids==x,,drop=FALSE])},integer(1))
    if(sum(cells_per_sample < MIN_CELLS) > 0){
        small_samples <- names(cells_per_sample)[cells_per_sample < MIN_CELLS]
        samples_to_warn <- paste(shQuote(small_samples, type="cmd"), collapse=", ")
        warning("The following samples have fewer than ", MIN_CELLS,
                " cells. This may lead to unreliable results.\n",
                samples_to_warn)
    }

    # If a sample has fewer cells than the `k` argument and `dens=="KNN"`,
    # its cells are ignored and its pairwise divergences are NA-valued.
    knn_na_values <- FALSE # Flag to add NA divergences to output matrix
    if(dens == "KNN" && (sum(cells_per_sample < k) > 0)){
        knn_na_values <- TRUE
        knn_withheld_samples <- names(cells_per_sample)[cells_per_sample < k]
        samples_to_warn <- paste(shQuote(knn_withheld_samples, type="cmd"), collapse=", ")
        warning("The following samples have fewer than the minimum of ", k,
                " cells required to run k-Nearest Neighbors. ",
                "Divergence pairs including these samples have value NA in the divergence matrix.\n",
                samples_to_warn)
        # Remove cells from the affected units
        embedding_matrix <- embedding_matrix[!(cell_sample_ids %in% knn_withheld_samples),,drop=FALSE]
        cell_sample_ids <- cell_sample_ids[!(cell_sample_ids %in% knn_withheld_samples)]
        # Remove the associated sample IDs to avoid an error when running kNN
        unique_sample_ids <- unique(cell_sample_ids)
        names(unique_sample_ids) <- unique_sample_ids
    }

    # We create a list indexed by sample ID containing each sample's embedding
    # matrix
    sample_matrix_list <- lapply(unique_sample_ids,
        function(x){embedding_matrix[(cell_sample_ids==x),,drop=FALSE]})
    # Saved `mclust` densities can be used instead of running the package by
    # setting the `prefit_density` optional argument to a list indexed by sample ID
    # containing a fit `densityMclust` object for each
    if(is.null(prefit_density)){
        mod_list <- .calc_dens(sample_matrix_list, dens = dens, 
        k = k, num_components = num_components,
        GMM_params = GMM_params, BPPARAM = BPPARAM)
    } else {
        # Validate that prefit_density contains densityMclust objects
        for(i in seq_along(prefit_density)){
            obj <- prefit_density[[i]]
            obj_name <- names(prefit_density)[i]
            if(is.null(obj_name)) obj_name <- paste0("[[", i, "]]")
           
            # Check if it's a densityMclust object
            if(!inherits(obj, "densityMclust")){
                # Provide specific guidance if it's a Mclust object
                if(inherits(obj, "Mclust")){
                    stop(
                        sprintf(
                            "prefit_density%s is a 'Mclust' object, but 'densityMclust' is required.\n",
                            obj_name
                        ),
                        "  Use mclust::densityMclust() instead of mclust::Mclust().\n",
                        "  See ?gloscope for more details."
                    )
                } else {
                    stop(
                        sprintf(
                            "prefit_density%s must be a 'densityMclust' object.\n",
                            obj_name
                        ),
                        "  Found object of class: ", class(obj)[1], "\n",
                        "  Use mclust::densityMclust() to create the correct object type.\n"
                    )
                }
            }
        }
        mod_list <- prefit_density
    }


    sample_pairs <- utils::combn(unique_sample_ids, 2)
    # Convert patient pairs to a list for BiocParallel::bplapply
    patient_pair_list <- lapply(seq_len(ncol(sample_pairs)), function(i) sample_pairs[,i])
    # IMPORTANT: There are additional algorithms for divergence estimation
    # implemented in this package which are not accessible from the `gloscope`
    # function. The optional arguments `varapp`, `epapp`, and `ep` must be manually
    # set below. See `R/.calc_dist.R` for their details.
    divergence_list <- BiocParallel::bplapply(patient_pair_list,
        function(w){ .calc_dist(mod_list = mod_list, s1 = w[1], s2 = w[2],
            df_list = sample_matrix_list, dist_metric = dist_metric, dens = dens,
            r = r, k = k, KNN_params = KNN_params,
            varapp = FALSE, epapp = FALSE, ep = NA)},BPPARAM=BPPARAM)
    
    # Convert pair-wise distances to a symmetric distance matrix
    divergence_vec <- unlist(divergence_list)
    divergence_matrix <- matrix(0, ncol = length(unique_sample_ids),
                                nrow = length(unique_sample_ids))
    rownames(divergence_matrix) <- unique_sample_ids
    colnames(divergence_matrix) <- unique_sample_ids
    divergence_matrix[lower.tri(divergence_matrix, diag=FALSE)] <- divergence_vec
    divergence_matrix[upper.tri(divergence_matrix)] <- t(divergence_matrix)[upper.tri(divergence_matrix)]

    if(knn_na_values){
        # pad matrix with NA divergences if kNN density estimate cannot be run for
        # some units
        full_samples <- c(unique_sample_ids, knn_withheld_samples)
        num_total <- length(full_samples)
        num_included <- length(unique_sample_ids)
        num_withheld <- length(knn_withheld_samples)
        na_pad_col <- matrix(NA, nrow = num_included, ncol = num_withheld)
        na_pad_row <- matrix(NA, nrow = num_withheld, ncol = num_total)
        divergence_matrix <- cbind(divergence_matrix,na_pad_col)
        divergence_matrix <- rbind(divergence_matrix,na_pad_row)
        rownames(divergence_matrix) <- full_samples
        colnames(divergence_matrix) <- full_samples
    }

    if(dens == "GMM"){
        mod_list <- lapply(mod_list,
            function(x){x[-which(names(x) %in% c("data", "classification","uncertainty", "density"))]})
    }
    if(return_density && dens == "GMM"){
        return(list(dist = divergence_matrix, modlist = mod_list))
    } else{
        return(divergence_matrix)
    }
}

library(testthat)
library(mclust)

test_that("gloscope validates prefit_density object types", {
    # Use the small test data from setup_data.R
    sample_ids <- subsample_metadata$sample_id
    
    # Create sample embedding list for testing (needed for creating mclust objects)
    embeddings_list <- lapply(unique(sample_ids), function(x) {
        subsample_data_subset[(sample_ids == x), ]
    })
    names(embeddings_list) <- unique(sample_ids)
    
    # Test 1: Error when prefit_density contains Mclust object (wrong type)
    mclust_obj <- mclust::Mclust(
        embeddings_list[[1]], 
        G = 5, 
        modelNames = "VVE",
        verbose = FALSE
    )
    prefit_wrong_mclust <- list(mclust_obj)
    names(prefit_wrong_mclust) <- names(embeddings_list)[1]
    
    expect_error(
        gloscope(
            subsample_data_subset,
            sample_ids,
            dens = "GMM",
            prefit_density = prefit_wrong_mclust
        ),
        regexp = "is a 'Mclust' object, but 'densityMclust' is required"
    )
    
    # Test 2: Error message suggests correct function (densityMclust)
    expect_error(
        gloscope(
            subsample_data_subset,
            sample_ids,
            dens = "GMM",
            prefit_density = prefit_wrong_mclust
        ),
        regexp = "mclust::densityMclust\\(\\)"
    )
    
    # Test 3: Error when prefit_density contains completely wrong object type
    prefit_wrong_type <- list(matrix(1:10, nrow = 5))
    names(prefit_wrong_type) <- names(embeddings_list)[1]
    
    expect_error(
        gloscope(
            subsample_data_subset,
            sample_ids,
            dens = "GMM",
            prefit_density = prefit_wrong_type
        ),
        regexp = "must be a 'densityMclust' object"
    )
    
    # Test 4: Error message includes object name for named list
    expect_error(
        gloscope(
            subsample_data_subset,
            sample_ids,
            dens = "GMM",
            prefit_density = prefit_wrong_mclust
        ),
        regexp = names(embeddings_list)[1]
    )
    
    # Test 5: Error message includes index [[1]] for unnamed list
    prefit_unnamed <- list(mclust_obj)
    expect_error(
        gloscope(
            subsample_data_subset,
            sample_ids,
            dens = "GMM",
            prefit_density = prefit_unnamed
        ),
        regexp = "\\[\\[1\\]\\]"
    )
    
    # Test 6: Success with correct densityMclust objects
    # Create correct densityMclust objects for all samples
    prefit_correct <- lapply(names(embeddings_list), function(sample_name) {
        mclust::densityMclust(
            embeddings_list[[sample_name]],
            G = 5,
            modelNames = "VVE",
            plot = FALSE,
            verbose = FALSE
        )
    })
    names(prefit_correct) <- names(embeddings_list)
    
    expect_silent(
        result <- gloscope(
            subsample_data_subset,
            sample_ids,
            dens = "GMM",
            prefit_density = prefit_correct,
            BPPARAM = BiocParallel::SerialParam(RNGseed = 2)
        )
    )
    
    # Verify result is a valid divergence matrix
    expect_equal(dim(result), c(length(unique(sample_ids)), length(unique(sample_ids))))
    expect_equal(isSymmetric(result), TRUE)
    expect_equal(unname(diag(result)), rep(0, length(unique(sample_ids))))
})

# The examples, test cases, and vignette for this package use a subset of
# peripheral blood mononuclear cells (PBMCs) scRNA-Seq data from
# COVID-infected patients and healthy controls presented in
# "Single-cell multi-omics analysis of the immune response in COVID-19"
# by Stephenson et al. (2021) [doi: 10.1038/s41591-021-01329-2]. This 
# script recreates the two .Rda files found in the `data/` folder.

run_script <- FALSE
if(run_script){
# An `.h5ad` file containing the processed counts in Seurat format is
# downloaded from the European Bioinformatics Institute with
# Accession E-MTAB-10026. Note that this file is on the order of 8 GB
# and can take several hours to download via FTP.

here::i_am("inst/script/create_example_data.R")
library(SingleCellExperiment)
library(zellkonverter)

# Execute a shell script provided by the European Bioinformatics Institute
system2(command="bash", args="inst/script/ftp_data.sh")

# Load .h5ad file as SingleCellExperiment
full_sce <- zellkonverter::readH5AD(here::here("inst","extdata","covid_portal_210320_with_raw.h5ad"))

# Subset to 500 cells from 20 patients
subset_cell_labels <- readRDS(here::here("inst","extdata","subset_cell_labels.Rds"))
subset_sce <- full_sce[,subset_cell_labels]

# Create a minimal SCE with only metadata and PCA embeddings
subset_metadata <- SingleCellExperiment::colData(subset_sce)[,c("patient_id","Status","initial_clustering")]
colnames(subset_metadata) <- c("sample_id","phenotype","cluster_id")
subset_metadata[] <- lapply(subset_metadata, as.factor)

subset_pca <- SingleCellExperiment::reducedDim(subset_sce,"X_pca")
pca_colnames <- paste0(rep("PC_",50),1:50)
colnames(subset_pca) <- pca_colnames
zero_counts <- matrix(0,2,10000) # Necessary to add PCA to SCE
colnames(zero_counts) <- subset_cell_labels

example_SCE <- SingleCellExperiment::SingleCellExperiment(assays=list(zero_counts), reducedDims=list(PCA=subset_pca))
colData(example_SCE) <- subset_metadata
example_SCE@assays@data@listData <- list() # Remove zero count matrix

save(example_SCE,file=here::here("inst","extdata","example_SCE.rda"))
}

#' @title Compute statistical divergences between GloScope representations of
#'   samples
#'
#' @description The `.calc_dist` function calculates a statistical divergences
#'   between two specified samples of an input list of sample GloScope
#'   representations. Different subroutines are called depending on the
#'   divergence specified (symmetric KL or Jensen-Shannon) and density
#'   estimation used for the GloScope representations (GMM or KNN).
#'
#' @param mod_list A named list with each sample's estimated density
#' @param s1 The name or index of the first sample in a pair (must be a key in
#'   mod_list)
#' @param s2 The name or index of the second sample in a pair
#' @param df_list A named list with each sample's reduced dimension embedding
#' @param dist_metric The distance metric to use (KL or JS)
#' @param dens The density estimation method (GMM or KNN)
#' @param r Number of Monte Carlo simulations to generate
#' @param k Number of k nearest neighbours for KNN density estimation, default k
#'   = 50.
#' @param KNN_params a list of arguments for either `FNN::KL.dist` (KL) or `RANN::nn2` (JS)
#' @param varapp Boolean for using variation approximation of KL divergence;
#'   NOTE: Currently disabled
#' @param epapp Boolean for applying an epsilon perturbation to MC calculated
#'   KL, default = FALSE
#' @param ep Epsilon perturbation size to add to MC KL divergence calculation,
#'   default = NA
#' @return The estimated statistical divergence between two GloScope
#'   represenations
#' @noRd
.calc_dist <- function(mod_list, s1, s2, df_list,
                dist_metric = c("KL","JS"), dens = c("GMM","KNN"), r, k,
                KNN_params, varapp = FALSE, epapp = FALSE, ep = NA){
    dens<-match.arg(dens)
    dist_metric<-match.arg(dist_metric)
    if(dist_metric == "KL"){
        if(dens == "KNN"){
        if (typeof(KNN_params) == "list" & "k" %in% names(KNN_params)){
            stop("k cannot be specified in `KNN_params`. This is specified by `k` instead.")
        }
        KNN_params <- c(list(k=k),KNN_params)
            mydist <- .calc_kl(mod_list = mod_list, df_list = df_list, sample1 = s1, sample2 = s2,
                dens = dens, KNN_params = KNN_params)
        } else{
            # symmeterize by hand
            mydist <- .calc_kl(mod_list = mod_list, df_list = df_list, sample1 = s1, sample2 = s2,
                dens = dens, r = r, varapp = varapp, epapp = epapp, ep = ep) +
                .calc_kl(mod_list = mod_list, df_list = df_list, sample1 = s2, sample2 = s1,
                    dens = dens, r = r, varapp = varapp, epapp = epapp, ep = ep)
        }
    }  else if(dist_metric == "JS"){
        mydist <- .calc_JS(mod_list = mod_list, df_list = df_list, sample1 = s1, sample2 = s2,
            dens = dens, r = r, KNN_params = KNN_params)
    }

    return(mydist)
}

#' @title Calculate the KL divergence between a single pair of samples
#'
#' @description The `.calc_kl` function calculates the symmetric KL divergence
#'   between two GloScope representations. This is implemented with Monte Carlo
#'   approximation (with an optional epsilon perturbation term) if GMM is used
#'   for the density estimate of each cell or a plug-in formula if KNN is used
#'   for the density.
#'
#' @param df_list A named list with each sample's reduced dimension embedding
#' @param mod_list A named list with each sample's estimated density
#' @param sample1 The name or index of the first sample in a pair (must be a key
#'   in mod_list)
#' @param sample2 The name or index of the second sample in a pair
#' @param dens The density estimation method (GMM or KNN)
#' @param r Number of Monte Carlo simulations to generate
#' @param KNN_params a list of arguments to the FNN:KL.dist function
#' @param varapp Boolean for using variation approximation of KL divergence;
#'   NOTE: Currently disabled'
#' @param epapp Boolean for applying an epsilon perturbation to MC calculated
#'   KL, default = FALSE
#' @param ep Epsilon perturbation size to add to MC KL divergence calculation,
#'   default = NA
#' @return a numeric value of distance between sample1 and sample2's
#'   distribution.
#' @importFrom FNN KL.dist
#' @importFrom stats predict
#' @noRd
.calc_kl <- function(mod_list, df_list, sample1, sample2, dens, r = 10000 ,
            KNN_params = NULL, varapp=FALSE, epapp = FALSE, ep = NA){
    if(dens == "GMM"){
        mclust_mod1 <- mod_list[[sample1]]
        mclust_mod2 <- mod_list[[sample2]]
        s <- .sample_mclust(mclust_mod1, r=r)
        pi_1 <- mclust_mod1$parameters$pro
        pi_2 <- mclust_mod2$parameters$pro
        cov_1 <- mclust_mod1$parameters$variance$sigma
        cov_2 <- mclust_mod2$parameters$variance$sigma

        mu_1 <- mclust_mod1$parameters$mean
        mu_2 <- mclust_mod2$parameters$mean

        # Deprecated option for approximating KL
        if(varapp) {
            stop("A variational approximation to the KL divergence is not available at this time")
        } else {
            dens1 <- stats::predict(mclust_mod1, s, what = "dens", logarithm = TRUE)
            dens2 <- stats::predict(mclust_mod2, s, what = "dens", logarithm = TRUE)
            if(epapp) {
                kl <- sum(dens1 - (dens2+ep))/r
            } else {
                kl <- sum(dens1 - dens2) / r
            }
        }
    } else if(dens == "KNN"){
        KNN_params$X <- as.matrix(mod_list[[sample1]])
        KNN_params$Y <- as.matrix(mod_list[[sample2]])
        kl <- do.call(FNN::KL.dist,KNN_params)[KNN_params$k]
    }
    return(kl)
}

#' @title Calculate the Jensen-Shannon distance between a single pair of samples
#'
#' @description The `.calc_JS` function calculates the Jensen-Shannon distance
#'   between two GloScope representations. This is implemented with Monte Carlo
#'   approximation (with an optional epsilon perturbation term) if GMM is used
#'   for the density estimate of each cell or a plug-in formula if KNN is used
#'   for the density.
#'
#' @param df_list A named list with each sample's reduced dimension embedding
#' @param mod_list A named list with each sample's estimated density
#' @param sample1 The name or index of the first sample in a pair (must be a key
#'   in mod_list)
#' @param sample2 The name or index of the second sample in a pair
#' @param dens The density estimation method (GMM or KNN)
#' @param r Number of Monte Carlo simulations to generate
#' @param KNN_params a list of arguments for the `RANN::nn2` function
#' @return a numeric value of distance between sample1 and sample2's
#'   distribution.
#' @importFrom stats predict
#' @noRd
.calc_JS <- function(mod_list, df_list, sample1, sample2, dens,
            r , KNN_params){
    if(dens == "GMM"){
        mclust_mod1 <- mod_list[[sample1]]
        mclust_mod2 <- mod_list[[sample2]]
        s1 <- .sample_mclust(mclust_mod1, r=r)
        s2 <- .sample_mclust(mclust_mod2, r=r)
        dens1_1 <- stats::predict(mclust_mod1, s1, what = "dens", logarithm = TRUE)
        dens2_1 <- stats::predict(mclust_mod2, s1, what = "dens", logarithm = TRUE)
        mixture_1 <- log(1/2*exp(dens1_1) + 1/2*exp(dens2_1))
        dens1_2 <- stats::predict(mclust_mod1, s2, what = "dens", logarithm = TRUE)
        dens2_2 <- stats::predict(mclust_mod2, s2, what = "dens", logarithm = TRUE)
        mixture_2 <- log(1/2*exp(dens1_2) + 1/2*exp(dens2_2))
        js <- sum(dens1_1 - mixture_1)/(2*r) + sum(dens2_2 - mixture_2)/(2*r)
    }else if(dens == "KNN"){
        k <- KNN_params$k
        knn1_1 <- .knn_query(df_list, input = sample1, query = sample1, KNN_params = KNN_params)
        knn2_2 <- .knn_query(df_list, input = sample2, query = sample2, KNN_params = KNN_params)
        knn2_1 <- .knn_query(df_list, input = sample2, query = sample1, KNN_params = KNN_params)
        knn1_2 <- .knn_query(df_list, input = sample1, query = sample2, KNN_params = KNN_params)
        knn1 <- mod_list[[sample1]]
        knn2 <- mod_list[[sample2]]
        if(dim(knn1)[2] != dim(knn2)[2]){stop("This method assumes both densities are of equal dimension.")}
        ndim <- dim(knn1)[2]
        js <- 1/(2*dim(knn1)[1])*sum(log(2*dim(knn2)[1] * knn2_1^ndim/(dim(knn2)[1] * knn2_1^ndim +
                (dim(knn1)[1]-1) * knn1_1^ndim))) +
            1/(2*dim(knn2)[1])*sum(log(2*dim(knn1)[1] * knn1_2^ndim/(dim(knn1)[1] * knn1_2^ndim +
                (dim(knn2)[1]-1) * knn2_2^ndim)))
    }
    return(js)
}

#!/bin/bash
# Run this file in bash with this command:  ./filename
cd inst/extdata
HOST=ftp.ebi.ac.uk
USER=anonymous
ftp -pinv $HOST <<EOF
user $USER
cd biostudies/nfs/E-MTAB-/026/E-MTAB-10026/Files
binary
mget "covid_portal_210320_with_raw.h5ad"
disconnect
bye
EOF

---
title: "scRNA-Seq Population-level Analysis using GloScope"
author:
- name: Hao Wang
  affiliation:
  - &UCB University of California, Berkeley, California, USA
  email: hao_wang@berkeley.edu
- name: William Torous
  affiliation:
  - *UCB
- name: Elizabeth Purdom
  affiliation:
  - *UCB
date: "`r Sys.Date()`"
output: 
  BiocStyle::html_document:
    toc: true
bibliography: gloscope.bib
vignette: >
  %\VignetteIndexEntry{GloScope}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
```


# Introduction 

This vignette will review the steps needed to implement the `GloScope` methodology. `GloScope` is a framework for creating profiles of scRNA-Seq samples in order to globally compare and analyze them across patients or tissue samples. First this methodology estimates a global gene expression distribution for each sample, and then it calculates how divergent pairs of samples are from each other. The output from the package's main function, `gloscope()`, is a $n\times n$ divergence matrix containing the pairwise statistical divergences between all samples. This divergence matrix can be the input to other downstream statistical and machine learning tools. This package has been submitted to Bioconductor because we believe it may be of particular interest to the bioinformatic community and because it depends on other packages from Bioconductor.

`GloScope` estimates the gene expression density from a low dimensional embedding of the UMI count measurements, such as PCA or scVI embeddings. Users must provide `GloScope` with a `data.frame` containing each cell's embedding coordinates, along with the metadata which identifies to which sample each cell belongs to.

The package provides a helper function `plotMDS` which allows the user to visualize the divergence matrix with multidimensional scaling (MDS), but the divergence matrix can also be visualized with other algorithms as well.  

A standard workflow for `GloScope` consists of:

1. Obtain the dimension reduction embedding of the cells and specify how many dimensions to keep. This is computed outside of the `GloScope` package.

2. Choose a density estimation method (Gaussian mixture or k-nearest neighbours) to estimate each sample's latent distribution.

3. Calculate the symmetric KL divergence or Jensen-Shannon divergence between all pairs of samples.

4. Visualize the distance matrix with the first two dimensions of MDS using the `plotMDS` function.

## Installation

You can install the latest stable release of `GloScope` from Bioconductor. Make sure that you have the `BiocManager` package installed to proceed.

```{r install_bioc, eval=FALSE}
if (!require("BiocManager", quietly = TRUE)) {
    install.packages("BiocManager")
}
BiocManager::install("GloScope")
```

# Example

In this section, we use a toy example to illustrate the input, output and visualization steps of the `GloScope` pipeline. 

The data is a subset of that presented in @stephenson2021single, and was obtained from [this hyperlinked URL](https://www.ebi.ac.uk/biostudies/arrayexpress/studies/E-MTAB-10026). The dataset has a total of 647,366 peripheral blood mononuclear cells (PBMCs) from 130 patients, whose phenotypes are COVID-infected, healthy control donor, patients with other non-COVID respiratory disease, and volunteers administered with intravenous lipopolysaccharide (IV-LPS). To enable faster computation in this tutorial, we subset this dataset to 20 random COVID and healthy control donors and further subsample each patient's count matrix to  500 random cells. This subsampled data is provided in the SingleCellExperiment object `example_SCE`. We emphasize that this subsampling procedure is only for demonstration purposes and is not a recommended step in normal analyses.

## Data Input

We first load the `GloScope` package and the aforementioned example dataset.

```{r load-libs, message = FALSE,  warning = FALSE}
library(GloScope)
data("example_SCE")
```

The example SingleCellExperiment object contains the first $50$ principal components of the subsampled cells, as well as the sample and phenotype labels associated with each cell. 

```{r}
head(SingleCellExperiment::reducedDim(example_SCE,"PCA")[,1:10])
head(SingleCellExperiment::colData(example_SCE))
```

The following table confirms that each donor provides $500$ cells, and that both phenotypes are represented.

```{r}
table(SingleCellExperiment::colData(example_SCE)$sample_id, SingleCellExperiment::colData(example_SCE)$phenotype)
```

`GloScope` expects that the user provides `GloScope` with a `data.frame` containing each cell's low-dimensional embedding (with cells in rows), along with a vector which contains the sample from which each cell in the embedding matrix is drawn. The PCA embeddings in the example dataset are provided by the authors of @stephenson2021single. In general, users of `GloScope` can input any dimensionality reduction to the method. UMI counts are often provided as `Seurat` or `SingleCellExperiment` objects, and many dimensionality reduction strategies, including PCA and scVI (@lopez2018deep) can be computed and saved within those data structures. This is recommended, as it will allow the user to keep the counts, embeddings, and the meta data (like the sample from which each cell was isolated) in the same structure and will minimize the change of inadvertently mislabelling the sample of a cell.

The following code, which is only pseudo-code and not evaluated (since we do not provide the `seurat_object`), demonstrates how PCA embeddings and sample labels can be extracted from a `Seurat` object for input into the `gloscope` function.

```{r eval = F}
embedding_df <- seurat_object@reductions$pca@cell.embeddings
sample_ids <- seurat_object@meta.data$sample_id
```

## Divergence Matrix Calculation

With the cell embeddings and the sample labels in the proper format, the `gloscope` function is simple to setup and run. For simplicity, we will save these as separate objects, though for real datasets, this would not be recommended since it would unnecessarily make copies of the data and increase memory usage:

```{r}
embedding_matrix <- SingleCellExperiment::reducedDim(example_SCE,"PCA")[,1:10]
sample_ids <- SingleCellExperiment::colData(example_SCE)$sample_id
```

Although the example data contains the first 50 principal components, we chose to use only the first 10 for calculations. Large number of latent variables will make the density estimation unstable, so we do not recommend large increases to the number of latent variables. 

The base function call is `gloscope(embedding_df, sample_ids)`.  

```{r, eval=FALSE}
# Can take a couple of minutes to run:
gmm_divergence <- gloscope(embedding_matrix, sample_ids)
```

The default implementation, run above, implements the `GMM` option for density estimation; this is the method primarily considered in the manuscript, but can take longer to run so we haven't evaluated it here. (You can evaluate it by changing `eval=FALSE` to `eval=TRUE`).  

An alternative methods uses a non-parametric alternative for density estimation based on a $k$-nearest neighbours algorithm and can be chosen with the argument `dens`. We will use this on our examples simply to make the tutorial run quickly:

```{r}
knn_divergence <- gloscope(embedding_matrix, sample_ids, dens="KNN")
knn_divergence[1:5,1:5]
```


Note, that unlike PCA, not every dimensionality reduction method retains its statistical properties when only a subset of the coordinates is retained, with scVI being an example. For methods like scVI, you should choose the number of latent variables you will want to use *when calculating the latent variables* and not subset them after the fact. 

## More on the density method:

If the user chooses the default GMM method (`dens="GMM"`), `gloscope` fits sample-level densities with Gaussian mixture models (GMMs) implemented by `mclust`. The `mclust` package uses the Bayesian information criterion (BIC) to select a GMM from a family of models, and the user of `GloScope` can specify how many centroids should be considered in that family. By default GMMs with $5,10,15,20$ centroids are compared. Note that the previous default was $1,\dots,9$ centroids, with this change being implemented in development version 1.7. With the `num_components` optional vector to `gloscope` the user can specify the possible number of centroids for `mclust` to compute, with the one with the best BIC value being the final choice. 

The default Gaussian mixture model fit in GloScope is `VVE` of `mclust`. The user can set any optional arguments to control the `mclust::densityMclust` used for estimation through the `GMM_params` optional argument in the `gloscope` function. For example, `GMM_params = list(modelNames=c("VVV","VVE"),plot=TRUE)`, would fit both a `VVV` and `VVE` model for each sample and then return a single model chosen via BIC. The `plot` Boolean would cause `mclust` density plots to be printed during compilation.

When GMMs are used for density fitting, a Monte-Carlo approximation is then used to compute the pairwise divergences from the estimated densities. This means that the resulting estimate is stochastic, and details about controlling the random seed appear later in this vignette. The number of Monte-Carlo draws from the density of each sample is $10,000$ by default, and this is controlled by the optional parameter $r$ in the `gloscope` function.

```{r, eval=FALSE}
# Can take a couple of minutes to run:
gmm_divergence_alt<-gloscope(embedding_matrix, sample_ids, dens = "GMM",  num_components = c(2,4,6),r=20000)
```

A non-parametric alternative for density and divergence estimation is the $k$-nearest neighbours algorithm. To use this technique, the optional argument `dens="KNN"` should be set. The number of neighbors is a hyperparameter, equal to $50$ by default, and governed by the optional argument `k`. It is important to note that negative divergences between similar samples are possible with this density estimation choice. The `gloscope` function does not censor or round any negative values in the output matrix, leaving that decision to the user. Analgous to the optional argument `GMM_params` above, the user may specify an optional list `KNN_params` in the `gloscope` function call. This list will pass through optional arguments for either `FNN::KL.dist` (for KL) or `RANN::nn2` ( for JS).

```{r}
knn_divergence_alt <- gloscope(embedding_matrix, sample_ids, 
    dens = "KNN", k = 25)
```

## More on the divergence method

The default divergence for `GloScope` is the symmetric KL divergence, but the Jensen-Shannon divergence is also implemented. This can be controlled by setting the argument `dist_metric="KL"` or `dist_metric="JS"`, respectively. One beneficial property of the Jensen-Shannon divergence is that its square root is a proper distance metric. Note that `gloscope` returns a matrix of untransformed divergences, and the user must take the square root of matrix entries themselves if this is desired.


# Visualization

The `plotMDS` function provided by this package visualizes the output divergence matrix with multidimensional scaling. The `plotMDS` function utilizes the `isoMDS` function from the package `MASS`, and then creates a scatter plot with samples color-coded by a user-specified covariate such as phenotype 

This function requires a `data.frame` with the relevant metadata at the sample level, rather than at the cell level. This can easily be obtained by applying the `unique` function from base R to the cell-level metadata `data.frame`. 

```{r}
pat_info <- as.data.frame(unique(SingleCellExperiment::colData(example_SCE)[,-c(3)]))
head(pat_info)
```

Here we plot the MDS representation with each sample color-coded by the `phenotype` variable. Note the function call returns both a matrix of MDS embeddings and a `ggplot` visualization of the first two dimensions. 

```{r}
mds_result <- plotMDS(dist_mat = knn_divergence, metadata = pat_info, "sample_id",color_by="phenotype", k = 2)
mds_result$plot
```

Another classical way to visualize a divergence matrix is with a heatmap. The following code demonstrates that the output of `gloscope` is easily used in plotting functions beyond the package.

```{r}
heatmap(knn_divergence)
```

We also provide a simple wrapper to the more powerful `pheatmap` function in the `pheatmap` package that can also plot additional annotation information with the heatmap:

```{r}
plotHeatmap(knn_divergence, metadata = pat_info, "sample_id",color_by="phenotype")
```

# Testing and Confidence intervals
The `GloScope` package also provides functions to do simple inference on the resulting distance matrices, both permutation tests and bootstrap confidence intervals. There are currently three different test statistics we calculate: `anosim` and `adonis2` implemented in the `vegan` package, and `silhouette` implemented in the `cluster` package. Our functions are only applicable for testing differences between a simple grouping factor and provided for the convenience of the user. The `vegan` package provides greater options for more complex inference for distance matrices. 

The function `getMetrics` calculates any combination of these three statistics and optionally performs a permutation test (if `permuteTest=TRUE`). We would note that for permutation tests, `adonis2` is recommended by the authors over `anosim` (see their accompanying help files). 

```{r}
permResults <- getMetrics(dist_mat = knn_divergence, metadata = pat_info, sample_id="sample_id", group_vars="phenotype", permuteTest = TRUE)
permResults
```

We also provide a wrapper function for bootstrap confidence intervals of these statistics with the function `bootCI` which has the same input arguments. We only recommend using `anosim` and `silhouette` since the statistic from `adonis2` is a F statistic, and less interpretable for confidence intervals. 

```{r}
bootCIResults <- bootCI(dist_mat = list("KNN k=50"=knn_divergence, "KNN k=25"=knn_divergence_alt), metadata = pat_info, metrics=c("anosim","silhouette"), sample_id="sample_id", group_vars="phenotype")
bootCIResults
```
`bootCI` allows the user to give a list of multiple distance matrices to make it easy to plot comparative confidence intervals. (This option of looping over multiple distances is not available for `getMetrics`). `plotCI` gives a function that will create a `ggplot` object for plotting the confidence intervals

```{r}
ppg<-plotCI(bootCIResults,color_by="distance",group_by="metric")
ppg
```

# Parallelization and Random Seeds

To speed-up calculations of the pair-wise divergences, `GloScope` allows for parallelizing the calculation. The argument `BPPARAM` controls the parameters of the parallelization (see `bplapply`). 

The default is no parallelization, but the iteration across sample-pairs will still via the function `bplapply`. In this case (i.e. no parallelization), the argument is simply `BPPARAM = BiocParallel::SerialParam()`. 

IMPORTANT: Due to the construction of the NAMESPACE file, it is essential that any setting of the BPPARAM optional argument uses the `BiocParallel::` namespace prefix. For example, `gloscope(...,BPPARAM = MulticoreParam()` will raise an error, and `gloscope(...,BPPARAM = BiocParallel::MulticoreParam()` should be used instead.

## Random seed 

The calculation of the KL divergence from the GMM density estimate uses Monte-Carlo approximation, and hence has to randomly sample from the estimated density. To set the seed for the pseudo-random number generator used in the simulation, the seed needs to be set within the argument to `BPPARAM` and **not** by a call to `set.seed` (see https://bioconductor.org/packages/release/bioc/vignettes/BiocParallel/inst/doc/Random_Numbers.html for more information). 

This is how the seed must be set, *even if there is no parallelization* chosen (the default), because the iteration over sample pairs is sent through `bplapply` function regardless, as noted above. Setting the seed outside the function via `set.seed` will not have an effect on the function. 

The following is an example of how to set the random seed when running the `GMM` option, using the default of no parallelization:

```{r eval = FALSE}
gmm_divergence <- gloscope(embedding_matrix, sample_ids, dens = "GMM", dist_metric = "KL",
    BPPARAM = BiocParallel::SerialParam(RNGseed = 2))
```

The same argument (`RNGseed`) can be added to other `BPPARAM` arguments to set the seed.

Note that the `KNN` estimation procedure does not have any Monte-Carlo approximation steps, and thus does not need to have a random seed.

# References


# SessionInfo {-}

```{r sessionInfo}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gloscope.R
\name{gloscope}
\alias{gloscope}
\title{Calculate statistical divergence between all sample pairs}
\usage{
gloscope(
  embedding_matrix,
  cell_sample_ids,
  dens = c("GMM", "KNN"),
  dist_metric = c("KL", "JS"),
  r = 10000,
  num_components = c(5, 10, 15, 20),
  k = 50,
  GMM_params = list(modelNames = c("VVE"), verbose = FALSE, plot = FALSE),
  KNN_params = NULL,
  BPPARAM = BiocParallel::SerialParam(),
  prefit_density = NULL,
  return_density = FALSE
)
}
\arguments{
\item{embedding_matrix}{a matrix or data.frame of latent embeddings with rows
corresponding to cells and columns to dimensions}

\item{cell_sample_ids}{a vector of the samples IDs each cell comes from. Length
must match the number of rows in `embedding_matrix`}

\item{dens}{the density estimation. One of c("GMM","KNN")}

\item{dist_metric}{distance metric to calculate the distance. One of
c("KL","JS")}

\item{r}{number of Monte Carlo simulations to generate}

\item{num_components}{a vector of integers for the number of components to
fit GMMs to, default is c(5,10,15,20)}

\item{k}{number of nearest neighbours for KNN density estimation, default k =
50.}

\item{GMM_params}{optional mclust parameters, default is to restrict the fit
model to only VVE}

\item{KNN_params}{optional arguments for either `FNN:KL.dist` (KL) or
`RANN::nn2` (JS), default is NULL}

\item{BPPARAM}{BiocParallel parameters, default is running in serial. Set
random seed with `RNGseed` argument}

\item{prefit_density}{a named list of pre-fit `densityMclust` objects for
each sample, default is NULL}

\item{return_density}{return the GMM parameter list or not (if applicable),
default is FALSE}
}
\value{
A matrix containing the pairwise divergence or distance between all
  pairs of samples
}
\description{
This function calculates a matrix of pairwise divergences
  between input samples of single cell data.
}
\examples{
# Bring in small example data of single cell embeddings
data(example_SCE_small)
sample_ids <- SingleCellExperiment::colData(example_SCE_small)$sample_id 
pca_embeddings <- SingleCellExperiment::reducedDim(example_SCE_small,"PCA")
# Run gloscope on first 10 PCA embeddings
# We use 'KNN' option for speed ('GMM' is slightly slower)
pca_embeddings_subset <- pca_embeddings[,seq_len(10)] # select the first 10 PCs
dist_result <- gloscope(pca_embeddings_subset, sample_ids,
         dens="KNN", BPPARAM = BiocParallel::SerialParam(RNGseed=2))
dist_result

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotMDS.R
\name{plotMDS}
\alias{plotMDS}
\title{Plot the multidimensional scaling of the GloScope represenation}
\usage{
plotMDS(dist_mat, metadata_df, sample_id, k = 10, color_by, shape_by)
}
\arguments{
\item{dist_mat}{The divergence matrix output of `gloscope()`. Should be a
symmetric, square matrix.}

\item{metadata_df}{A data frame contains each sample's metadata. Note this is
NOT at the cell-level, and should have the same number of rows as dist_mat.}

\item{sample_id}{The column name or index in metadata_df that contains the
sample ID. This is for ensuring alignment between the dist_mat and the
metadata_df. The rownames of dist_mat are expected to match the sample_id
values.}

\item{k}{Number of MDS dimension to generate, default = 10}

\item{color_by}{The column name or index in metadata_df that should be used
to color the points by. If missing all points will be the same color.}

\item{shape_by}{The column name or index in metadata_df that should be used
to determine the shape of the points. If missing all points will be the
same shape.}
}
\value{
A list containing the MDS embedding and plot of the distance matrix
\itemize{
  \item mds - A data.frame containing the MDS embedding, with the number of rows equal to the number of samples.
  \item plot - A ggplot object containing the plot object. `print` of the object will create a plot.
}
}
\description{
This function calculates the multidimensional scaling for a
  GloScope divergence matrix and returns a ggplot object that plots it.
}
\details{
The function calls \code{\link[MASS]{isoMDS}} from the MASS package,
  calculates the requested k coordinates of the MDS plot. It also creates a
  ggplot object that will plot the first two dimensions color or shape coded
  by the given variables in the metadata data frame.
}
\examples{
data(example_SCE_small)
sample_ids <- SingleCellExperiment::colData(example_SCE_small)$sample_id
# Run gloscope on first 10 PCA embeddings
# We use 'KNN' option for speed ('GMM' is slightly slower)
pca_embeddings <- SingleCellExperiment::reducedDim(example_SCE_small,"PCA")
pca_embeddings_subset <- pca_embeddings[,seq_len(10)] # select the first 10 PCs
dist_result <- gloscope(pca_embeddings_subset, sample_ids,
   dens="KNN",
   BPPARAM = BiocParallel::SerialParam(RNGseed=2))
# make a per-sample metadata
sample_metadata <- as.data.frame(unique(SingleCellExperiment::colData(example_SCE_small)[,c(1,2)]))
mds_result <- plotMDS(dist_mat = dist_result, metadata_df = sample_metadata ,
sample_id="sample_id", color_by="phenotype",k=2)
head(mds_result$mds)
require(ggplot2)
mds_result$plot
# Add additional ggplot2 components to adapt figure
mds_result$plot + theme_bw()  + scale_color_manual(values=alpha(c("red","blue"),0.5))
}
\seealso{
\code{\link[MASS]{isoMDS}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{example_SCE}
\alias{example_SCE}
\alias{example_SCE_small}
\title{SingleCellExperiment containing example inputs to GloScope}
\format{
A SingleCellExperiment object with metadata and PCA embeddings
}
\value{
A SingleCellExperiment object
}
\description{
`example_SCE` is a SingleCellExperiment object which
  contains PCA embeddings and metadata for PBMCs from 20 COVID-infected
  and healthy control patients. Each sample is reduced
  to a random subset of 500 cells, for a total of 10,000 cells. The 
  `colData` slot of the object contains the metadata for each cell, 
  its sample ID and phenotype. The dimensionality reductions slot 
  contains the first 50 PCs, and these embeddings are provided by
  the authors of "Single-cell multi-omics analysis of the immune
  response in COVID-19" (Stephenson et al., 2021; doi: 10.1038/s41591-021-01329-2).

`example_SCE_small` is a SingleCellExperiment with the same structure as
  `example_SCE`, but only containing data from the first five samples. This is a
  smaller set for examples.
}
\examples{
# Code to create the small SCE from the full sample
# Reduction to 5 samples demonstrates data extraction from SCE objects
data(example_SCE)
sample_ids <- SingleCellExperiment::colData(example_SCE)$sample_id
whKeep <- which(sample_ids \%in\% unique(sample_ids)[seq_len(5)])
example_SCE_small <- SingleCellExperiment::SingleCellExperiment(
 assays=list(counts=matrix(rep(0,2500),ncol=2500)),
 colData=SingleCellExperiment::colData(example_SCE)[whKeep,],
 reducedDims=list("PCA"=SingleCellExperiment::reducedDim(example_SCE,"PCA")[whKeep,]))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gloscope_proportion.R
\name{gloscopeProp}
\alias{gloscopeProp}
\title{Calculate divergence between all sample pairs' cell type proportion}
\usage{
gloscopeProp(
  cell_sample_ids,
  cell_type_ids,
  ep = 0,
  dist_metric = c("KL", "JS", "TV")
)
}
\arguments{
\item{cell_sample_ids}{a vector of the samples IDs each cell comes from. Length
must match the number of element in `cell_type_ids`}

\item{cell_type_ids}{a vector of user defined cell type}

\item{ep}{an numeric value added to the summary counts. Default ep = 0 means
nothing will be added.}

\item{dist_metric}{metric to calculate the divergence between samples.}
}
\value{
clusprop_dist a symmetric matrix of divergences
}
\description{
This function calculates a matrix of pairwise divergences
  between input samples' cell type proportion.
}
\details{
Options for `dist_metric` are as follows: "KL" calculates the
  symmetric-KL divergence. "JS" calculates the Jenson-Shannon distance. "TV"
  calculates the total variation distance.
}
\examples{
# Bring in small example data of single cell embeddings
data(example_SCE_small)
sample_id <- SingleCellExperiment::colData(example_SCE_small)$sample_id 
cluster_id <- SingleCellExperiment::colData(example_SCE_small)$cluster_id 
dist_result <- gloscopeProp(sample_id, cluster_id, ep = 0.5, 
                                   dist_metric = "KL")
dist_result
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotCI.R
\name{plotCI}
\alias{plotCI}
\title{Plot confidence intervals}
\usage{
plotCI(ci_df, color_by, group_by, dodge_width = 0.5)
}
\arguments{
\item{ci_df}{A data frame contains each sample's metadata. Note this is
NOT at the cell-level.}

\item{color_by}{The column name or index in ci_df that should be used
to color the confidence intervals by.}

\item{group_by}{The column name or index in ci_df that should be used to
determine how to group the confidence intervals. If missing all confidence
intervals will be plotted in an order determined internally.}

\item{dodge_width}{value passed to `width` argument of
\code{\link[ggplot2]{position_dodge}} if `group_by` variable is given.
Controls separation between confidence intervals with the same grouping
value.}
}
\value{
A plot of sample-pair divergences with confidence intervals
}
\description{
This function creates a `ggplot` object that plots the
  confidence intervals created by `bootCI`
}
\examples{
data(example_SCE_small)
sample_ids <- SingleCellExperiment::colData(example_SCE_small)$sample_id
# Run gloscope on first 10 PCA embeddings
# We use 'KNN' option for speed ('GMM' is slightly slower)
pca_embeddings <- SingleCellExperiment::reducedDim(example_SCE_small,"PCA")
pca_embeddings_subset <- pca_embeddings[,seq_len(10)] # select the first 10 PCs
dist_result <- gloscope(pca_embeddings_subset, sample_ids,
   dens="KNN",
   BPPARAM = BiocParallel::SerialParam(RNGseed=2))
# make a per-sample metadata
sample_metadata <- as.data.frame(unique(SingleCellExperiment::colData(example_SCE_small)[,c(1,2)]))
# make another variable
sample_metadata$fakeGroup<-c(rep(c("A","B"),each=2),"A")
manyboot<-bootCI(dist_result,
  sample_metadata,"sample_id",
  metrics=c("anosim","silhouette"),group_vars=c("phenotype","fakeGroup"),R=20)
plotCI(manyboot,group_by="metric",color_by="grouping")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_metrics.R, R/bootCI.R
\name{getMetrics}
\alias{getMetrics}
\alias{bootCI}
\alias{bootGloscope}
\title{Calculate metrics on separation in distance matrix due to grouping variable}
\usage{
getMetrics(
  dist_mat,
  metadata_df,
  metrics = c("anosim", "adonis2", "silhouette"),
  sample_id,
  group_vars,
  checkData = TRUE,
  permuteTest = FALSE,
  permutations = 100
)

bootCI(
  dist_mat,
  metadata_df,
  metrics = "anosim",
  sample_id,
  group_vars,
  R = 1000,
  ci_type = c("perc", "norm", "basic", "stud", "bca"),
  ci_conf = 0.95,
  ...
)

bootGloscope(
  dist_mat,
  metadata_df,
  metrics = "anosim",
  sample_id,
  group_vars,
  R = 1000,
  ...
)
}
\arguments{
\item{dist_mat}{The divergence matrix output of `gloscope()`. Should be a
symmetric, square matrix. For `bootCI` the argument can be a list
of distance matrices.}

\item{metadata_df}{A data frame contains each sample's metadata. Note this is
NOT at the cell-level, and should have the same number of rows as dist_mat.}

\item{metrics}{vector of statistics to calculate. For `bootstrap_gloscope`
must be single value.}

\item{sample_id}{The column name or index in metadata_df that contains the
sample ID. This is for ensuring alignment between the dist_mat and the
metadata_df. The rownames of dist_mat are expected to match the sample_id
values.}

\item{group_vars}{vector of names of grouping variables in metadata_df for
which to calculate metrics. For `bootstrap_gloscope` must be single value.}

\item{checkData}{Whether to check whether dist_mat, metadata_df, and
sample_id match, for example in terms of dimensions and rownames. Mainly
used internally.}

\item{permuteTest}{whether to run permutation tests on each of the metrics}

\item{permutations}{if `permuteTest=TRUE`, an integer value defines the
number of permutations. Can also except output of
\code{\link[permute]{how}} for more fine control of the permutation
mechanisms.}

\item{R}{number of bootstrap replicates. See \code{\link[boot]{boot}}.}

\item{ci_type}{Single character value. The type of confidence interval to
compute. Passed to argument `type` in \code{\link[boot]{boot.ci}}}

\item{ci_conf}{Scalar value between 0 and 1. The confidence level requested.
Passed to argument `conf` in \code{\link[boot]{boot.ci}}.}

\item{...}{arguments passed to \code{\link[boot]{boot}}}
}
\value{
`getMetrics` creates a data frame containing the statistic for each
  combination of metric and grouping variable with columns
\itemize{
  \item metric 
  \item grouping
  \item statistic
  \item pval (if `permuteTest=TRUE`)
}

`bootCI` creates a data frame containing the statistic for
  each combination of metric and grouping variable with columns with the
  upper and lower bounds of the requested confidence intervals
\itemize{
  \item metric 
  \item grouping
  \item statistic
  \item lower
  \item upper
}

`bootGloscope` returns an object of class `boot` created by \code{\link[boot]{boot}}.
}
\description{
These functions are wrappers for calculating common metrics for
  the amount of separation in a distance matrix due to a grouping (factor)
  variable and creating bootstrap confidence intervals and permutation tests.
}
\details{
The function `getMetrics` is a simple wrapper for calculating statistics that
  summarize the difference between distances within and between groupings. If
  the variable defined by group_var does not have at least two groupings, the
  function will return a NA.

The options "anosim" and "adonis2" are wrappers to the functions of
  that name in the package `vegan`; we have turned off the permutation
  testing option of those functions. The functions in `vegan` have greater
  capability, and in particular \code{\link[vegan]{adonis2}} has capability
  to handle more complicated testing paradigms than a simple grouping factor.
  Permutation tests for these statistics are handled by the functions of
  `vegan`.

The option "silhouette" calls the function of that name from the
  package `cluster` and calculates the silhouette width of each group and
  then averages them across groups. The permutation test is coded making use
  of the package `permute`, similar to `vegan`, so that control of the
  permutation mechanism is possible in the same way.

`bootCI` is a wrapper function to
  \code{\link[boot]{boot.ci}}. `boot.ci` can be called directly on the output
  of `bootGloscope`. The main advantage of `bootCI` is to calculate
  bootstrap CI over multiple choices of metrics, variables, and/or distance
  matrices. Unlike `boot.ci`, `bootCI` does not allow different
  choices of confidence interval types or levels, so `ci_type` and `ci_level`
  must be of length 1. For this kind of multiplicity, call `boot.ci` directly
  on the output of `bootGloscope`.

The function `bootGloscope` is a wrapper to the
  \code{\link[boot]{boot}} function for creating bootstraps of one of the
  metrics calculated by `getMetrics`. Most users will probably prefer `bootCI`
}
\examples{
data(example_SCE_small)
sample_ids <- SingleCellExperiment::colData(example_SCE_small)$sample_id
# Run gloscope on first 10 PCA embeddings
# We use 'KNN' option for speed ('GMM' is slightly slower)
pca_embeddings <- SingleCellExperiment::reducedDim(example_SCE_small,"PCA")
pca_embeddings_subset <- pca_embeddings[,seq_len(10)] # select the first 10 PCs
dist_result <- gloscope(pca_embeddings_subset, sample_ids,
   dens="KNN",
   BPPARAM = BiocParallel::SerialParam(RNGseed=2))
# make a per-sample metadata
sample_metadata <- as.data.frame(unique(SingleCellExperiment::colData(example_SCE_small)[,c(1,2)]))
# make another variable
sample_metadata$grouping<-c(rep(c("A","B"),each=2),"A")
getMetrics(dist_result,metadata_df=sample_metadata, sample_id="sample_id", 
  group_vars="phenotype")
# run permutation tests:
getMetrics(dist_result,metadata_df=sample_metadata, sample_id="sample_id", 
  group_vars=c("phenotype","grouping"), permuteTest=TRUE)
  
# calculate many bootstraps -- for speed up we set R ridiculously low
manyboot<-bootCI(list("Distance 1"=dist_result,"Another distance"=dist_result),
  sample_metadata,"sample_id",
  metrics=c("anosim","silhouette"),group_vars=c("phenotype","grouping"),R=20)

# single bootstrap of anosim
bootout<-bootGloscope(dist_result,sample_metadata,"sample_id",
  metric="anosim",group_var="phenotype")
#work with the boot object using functions in boot package:
library(boot)
print(bootout)
boot.ci(bootout)

}
\seealso{
\code{\link[vegan]{anosim}}, \code{\link[vegan]{adonis2}}, 
 \code{\link[cluster]{silhouette}}, \code{\link[permute]{how}}

\code{\link[boot]{boot.ci}}

\code{\link[boot]{boot}}
}
